# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess

from pex.build_system.pep_518 import BuildSystem, load_build_system
from pex.pip.version import PipVersion
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.resolver_configuration import PipConfiguration, ReposConfiguration
from pex.targets import LocalInterpreter
from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_load_build_system_pyproject_custom_repos(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    current_target = LocalInterpreter.create()
    pip_version = (
        PipVersion.v22_2_2
        if PipVersion.v22_2_2.requires_python_applies(current_target)
        else PipVersion.VENDORED
    )
    build_system = load_build_system(
        current_target,
        ConfiguredResolver(PipConfiguration(version=pip_version)),
        pex_project_dir,
    )
    assert isinstance(build_system, BuildSystem)

    # Verify that we can still resolve a build backend even when our toml lock is unuseable.
    repository_pex = os.path.join(str(tmpdir), "repository.pex")
    run_pex_command(
        args=["--include-tools", "-o", repository_pex] + list(build_system.requires)
    ).assert_success()

    find_links = os.path.join(str(tmpdir), "find_links")
    subprocess.check_call(
        args=[repository_pex, "repository", "extract", "--find-links", find_links],
        env=make_env(PEX_TOOLS=1),
    )

    repos_configuration = ReposConfiguration.create(find_links=[find_links])
    assert not repos_configuration.indexes
    custom_resolver = ConfiguredResolver(
        PipConfiguration(repos_configuration=repos_configuration, version=pip_version)
    )
    build_system = load_build_system(current_target, custom_resolver, pex_project_dir)
    assert isinstance(build_system, BuildSystem)
    subprocess.check_call(
        args=[build_system.venv_pex.pex, "-c", "import {}".format(build_system.build_backend)]
    )



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
from argparse import ArgumentParser
from contextlib import contextmanager
from tempfile import NamedTemporaryFile

import pytest

from pex.bin.pex import (
    build_pex,
    configure_clp,
    configure_clp_pex_options,
    configure_clp_pex_resolution,
)
from pex.common import safe_copy, temporary_dir
from pex.compatibility import to_bytes
from pex.interpreter import PythonInterpreter
from pex.resolve import requirement_options, resolver_options, target_options
from pex.testing import (
    PY39,
    built_wheel,
    ensure_python_interpreter,
    run_pex_command,
    run_simple_pex,
)
from pex.typing import TYPE_CHECKING
from pex.venv.bin_path import BinPath

if TYPE_CHECKING:
    from typing import Iterator, List, Optional, Text


@contextmanager
def option_parser():
    # type: () -> Iterator[ArgumentParser]
    yield ArgumentParser()


# Make sure that we're doing append and not replace
def test_clp_requirements_txt():
    # type: () -> None
    parser = configure_clp()
    options = parser.parse_args(args="-r requirements1.txt -r requirements2.txt".split())
    assert options.requirement_files == ["requirements1.txt", "requirements2.txt"]


def test_clp_constraints_txt():
    # type: () -> None
    parser = configure_clp()
    options = parser.parse_args(args="--constraint requirements1.txt".split())
    assert options.constraint_files == ["requirements1.txt"]


def test_clp_arg_file():
    # type: () -> None
    with NamedTemporaryFile() as tmpfile:
        tmpfile.write(to_bytes("-r\nrequirements1.txt\r-r\nrequirements2.txt"))
        tmpfile.flush()

        parser = configure_clp()
        options = parser.parse_args(args=["@" + tmpfile.name])
        assert options.requirement_files == ["requirements1.txt", "requirements2.txt"]


def test_clp_preamble_file():
    # type: () -> None
    with NamedTemporaryFile() as tmpfile:
        tmpfile.write(to_bytes('print "foo!"'))
        tmpfile.flush()

        parser = configure_clp()
        options = parser.parse_args(args=["--preamble-file", tmpfile.name])
        assert options.preamble_file == tmpfile.name

        requirement_configuration = requirement_options.configure(options)
        resolver_configuration = resolver_options.configure(options)
        target_config = target_options.configure(options)
        targets = target_config.resolve_targets()
        pex_builder = build_pex(
            requirement_configuration=requirement_configuration,
            resolver_configuration=resolver_configuration,
            interpreter_constraints=target_config.interpreter_constraints,
            targets=targets,
            options=options,
        )
        assert pex_builder._preamble == 'print "foo!"'


def test_clp_prereleases():
    # type: () -> None
    with option_parser() as parser:
        configure_clp_pex_resolution(parser)

        options = parser.parse_args(args=[])
        assert not options.allow_prereleases

        options = parser.parse_args(args=["--no-pre"])
        assert not options.allow_prereleases

        options = parser.parse_args(args=["--pre"])
        assert options.allow_prereleases


def test_clp_prereleases_resolver():
    # type: () -> None
    with built_wheel(name="prerelease-dep", version="1.2.3b1") as prerelease_dep, built_wheel(
        name="transitive-dep", install_reqs=["prerelease-dep"]
    ) as transitive_dep, built_wheel(
        name="dep", install_reqs=["prerelease-dep>=1.2", "transitive-dep"]
    ) as dep, temporary_dir() as dist_dir, temporary_dir() as cache_dir:
        for dist in (prerelease_dep, transitive_dep, dep):
            safe_copy(dist, os.path.join(dist_dir, os.path.basename(dist)))

        parser = configure_clp()

        options = parser.parse_args(
            args=[
                # This test is run against all Pythons; so ensure we have a Pip that works with all
                # the pythons we support.
                "--pip-version=vendored",
                "--no-index",
                "--find-links",
                dist_dir,
                "--cache-dir",
                cache_dir,  # Avoid dangling {pex_root}.
                "--no-pre",
                "dep",
            ]
        )
        assert not options.allow_prereleases

        with pytest.raises(SystemExit):
            target_config = target_options.configure(options)
            build_pex(
                requirement_configuration=requirement_options.configure(options),
                resolver_configuration=resolver_options.configure(options),
                interpreter_constraints=target_config.interpreter_constraints,
                targets=target_config.resolve_targets(),
                options=options,
            )

        # When we specify `--pre`, allow_prereleases is True
        options = parser.parse_args(
            args=[
                # This test is run against all Pythons; so ensure we have a Pip that works with all
                # the pythons we support.
                "--pip-version=vendored",
                "--no-index",
                "--find-links",
                dist_dir,
                "--cache-dir",
                cache_dir,  # Avoid dangling {pex_root}.
                "--pre",
                "dep",
            ]
        )
        assert options.allow_prereleases

        # Without a corresponding fix in pex.py, this test failed for a dependency requirement of
        # dep==1.2.3b1 from one package and just dep (any version accepted) from another package.
        # The failure was an exit from build_pex() with the message:
        #
        # Could not satisfy all requirements for dep==1.2.3b1:
        #     dep==1.2.3b1, dep
        #
        # With a correct behavior the assert line is reached and pex_builder object created.
        target_config = target_options.configure(options)
        pex_builder = build_pex(
            requirement_configuration=requirement_options.configure(options),
            resolver_configuration=resolver_options.configure(options),
            interpreter_constraints=target_config.interpreter_constraints,
            targets=target_config.resolve_targets(),
            options=options,
        )
        assert pex_builder is not None
        assert len(pex_builder.info.distributions) == 3, "Should have resolved deps"


def test_clp_pex_options():
    with option_parser() as parser:
        configure_clp_pex_options(parser)

        options = parser.parse_args(args=[])
        assert options.venv == False

        options = parser.parse_args(args=["--venv"])
        assert options.venv == BinPath.FALSE

        options = parser.parse_args(args=["--venv", "append"])
        assert options.venv == BinPath.APPEND

        options = parser.parse_args(args=["--venv", "prepend"])
        assert options.venv == BinPath.PREPEND


def test_build_pex():
    # type: () -> None
    with temporary_dir() as sandbox:
        pex_path = os.path.join(sandbox, "pex")
        results = run_pex_command(["ansicolors==1.1.8", "--output-file", pex_path])
        results.assert_success()
        stdout, returncode = run_simple_pex(
            pex=pex_path, args=["-c", 'import colors; print(" ".join(colors.COLORS))']
        )
        assert 0 == returncode
        assert b"black red green yellow blue magenta cyan white" == stdout.strip()


def test_run_pex():
    # type: () -> None

    def assert_run_pex(python=None, pex_args=None):
        # type: (Optional[str], Optional[List[str]]) -> List[Text]
        pex_args = list(pex_args) if pex_args else []
        results = run_pex_command(
            python=python,
            args=pex_args
            + ["ansicolors==1.1.8", "--", "-c", 'import colors; print(" ".join(colors.COLORS))'],
            quiet=True,
        )
        results.assert_success()
        assert "black red green yellow blue magenta cyan white" == results.output.strip()
        return results.error.splitlines()

    incompatible_platforms_warning_msg = (
        "WARNING: attempting to run PEX with incompatible platforms!"
    )

    assert incompatible_platforms_warning_msg not in assert_run_pex()
    assert incompatible_platforms_warning_msg not in assert_run_pex(pex_args=["--platform=current"])
    assert incompatible_platforms_warning_msg not in assert_run_pex(
        pex_args=["--platform={}".format(PythonInterpreter.get().platform)]
    )

    py39 = ensure_python_interpreter(PY39)
    stderr_lines = assert_run_pex(python=py39, pex_args=["--platform=macosx-10.13-x86_64-cp-37-m"])
    assert incompatible_platforms_warning_msg in stderr_lines



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
import sys
from textwrap import dedent

from colors import cyan

from pex.common import filter_pyc_files, safe_open
from pex.testing import IntegResults, make_env, run_pex_command
from pex.typing import TYPE_CHECKING
from pex.util import CacheHelper
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Any, Set


def run_pex_tools(*args):
    # type: (*str) -> IntegResults

    process = subprocess.Popen(
        args=[sys.executable, "-mpex.tools"] + list(args),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()
    return IntegResults(
        output=stdout.decode("utf-8"), error=stderr.decode("utf-8"), return_code=process.returncode
    )


def test_collisions(
    tmpdir,  # type: Any
    pex_bdist,  # type: str
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")

    collision_src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(collision_src, "will_not_collide_with_pex_module.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                def verb():
                  return 42
                """
            )
        )
    with safe_open(os.path.join(collision_src, "setup.cfg"), "w") as fp:
        fp.write(
            dedent(
                """\
                [metadata]
                name = collision
                version = 0.0.1

                [options]
                py_modules =
                    will_not_collide_with_pex_module
                
                [options.entry_points]
                # Although will_not_collide_with_pex_module does not collide with Pex, the 
                # generated bin/pex script will collide with the Pex pex script.
                console_scripts =
                    pex = will_not_collide_with_pex_module:verb
                """
            )
        )
    with safe_open(os.path.join(collision_src, "setup.py"), "w") as fp:
        fp.write("from setuptools import setup; setup()")

    collisions_pex = os.path.join(str(tmpdir), "collisions.pex")
    run_pex_command(
        args=[
            pex_bdist,
            collision_src,
            "-o",
            collisions_pex,
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
        ]
    ).assert_success()

    venv_dir = os.path.join(str(tmpdir), "collisions.venv")
    result = run_pex_tools(collisions_pex, "venv", venv_dir)
    result.assert_failure()
    assert (
        "Encountered collision populating {venv_dir} from PEX at {pex}:\n"
        "1. {venv_dir}/bin/pex was provided by:".format(venv_dir=venv_dir, pex=collisions_pex)
    ) in result.error, result.error

    result = run_pex_tools(collisions_pex, "venv", "--collisions-ok", "--force", venv_dir)
    result.assert_success()
    assert (
        "PEXWarning: Encountered collision populating {venv_dir} from PEX at {pex}:\n"
        "1. {venv_dir}/bin/pex was provided by:".format(venv_dir=venv_dir, pex=collisions_pex)
    ) in result.error, result.error
    assert 42 == subprocess.call(args=[Virtualenv(venv_dir=venv_dir).bin_path("pex")])


def test_collisions_mergeable_issue_1570(tmpdir):
    # type: (Any) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "opencensus==0.8.0",
            "opencensus_context==0.1.2",
            "-o",
            pex,
            "--resolver-version",
            "pip-2020-resolver",
        ]
    ).assert_success()

    venv_dir = os.path.join(str(tmpdir), "venv")
    run_pex_tools(pex, "venv", venv_dir).assert_success()

    venv = Virtualenv(venv_dir=venv_dir)
    _, stdout, _ = venv.interpreter.execute(
        args=[
            "-c",
            dedent(
                """\
                from __future__ import print_function
                
                import os

                import opencensus
                import opencensus.common


                print(os.path.realpath(opencensus.__file__))
                print(os.path.realpath(opencensus.common.__file__))
                """
            ),
        ]
    )
    assert [
        os.path.join(venv.site_packages_dir, "opencensus", "__init__.py"),
        os.path.join(venv.site_packages_dir, "opencensus", "common", "__init__.py"),
    ] == stdout.splitlines()


def test_scope_issue_1631(tmpdir):
    # type: (Any) -> None

    src_dir = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src_dir, "app.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from colors import cyan

                print(cyan("Colluphid: Cupitt or Dawkins?"))
                """
            )
        )

    pex_root = os.path.join(str(tmpdir), "pex_root")
    app_pex = os.path.join(str(tmpdir), "app.pex")
    run_pex_command(
        args=[
            "-D",
            src_dir,
            "-m" "app",
            "ansicolors==1.1.8",
            "--include-tools",
            "-o",
            app_pex,
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
        ]
    ).assert_success()

    def execute_venv_tool(
        venv_dir,  # type: str
        *args  # type: str
    ):
        # type: (...) -> None
        subprocess.check_call(
            args=[app_pex, "venv", venv_dir] + list(args), env=make_env(PEX_TOOLS=1)
        )

    def assert_app(venv_dir):
        # type: (str) -> Virtualenv
        assert (
            cyan("Colluphid: Cupitt or Dawkins?")
            == subprocess.check_output(args=[os.path.join(venv_dir, "pex")]).decode("utf-8").strip()
        )
        return Virtualenv(venv_dir)

    def recursive_listing(venv_dir):
        # type: (str) -> Set[str]
        return {
            os.path.relpath(os.path.join(root, f), venv_dir)
            for root, _, files in os.walk(venv_dir)
            for f in filter_pyc_files(files)
        }

    def site_packages_path(
        venv_dir,  # type: str
        *relpath  # type: str
    ):
        # type: (...) -> str
        venv = Virtualenv(venv_dir)
        return os.path.relpath(os.path.join(venv.site_packages_dir, *relpath), venv.venv_dir)

    def app_py_path(venv_dir):
        # type: (str) -> str
        return site_packages_path(venv_dir, "app.py")

    def colors_package_path(venv_dir):
        # type: (str) -> str
        return site_packages_path(venv_dir, "colors", "__init__.py")

    venv_directory = os.path.join(str(tmpdir), "venv")
    execute_venv_tool(venv_directory)
    canonical_venv_listing = recursive_listing(venv_directory)
    venv = assert_app(venv_directory)
    # N.B.: Some venv activation scripts in bin/ differ since they hard code the venv python
    # interpreter path in them. Since we have a full canonical_venv_listing to check, we settle on
    # testing contents are identical just for site-packages which is what is important anyhow: it's
    # where the deps and user code are installed to.
    canonical_venv_hash = CacheHelper.dir_hash(venv.site_packages_dir)

    # 1st populate dependencies, then sources.
    venv_directory = os.path.join(str(tmpdir), "venv.deps-srcs")
    execute_venv_tool(venv_directory, "--scope=deps")
    colors_package = colors_package_path(venv_directory)
    app_module = app_py_path(venv_directory)

    assert colors_package in recursive_listing(venv_directory)
    assert app_module not in recursive_listing(venv_directory)

    execute_venv_tool(venv_directory, "--scope=srcs")
    assert colors_package in recursive_listing(venv_directory)
    assert app_module in recursive_listing(venv_directory)

    assert canonical_venv_listing == recursive_listing(venv_directory)
    venv = assert_app(venv_directory)
    assert canonical_venv_hash == CacheHelper.dir_hash(venv.site_packages_dir)

    # 1st populate sources, then dependencies.
    venv_directory = os.path.join(str(tmpdir), "venv.srcs-deps")
    execute_venv_tool(venv_directory, "--scope=srcs")
    colors_package = colors_package_path(venv_directory)
    app_module = app_py_path(venv_directory)

    assert app_module in recursive_listing(venv_directory)
    assert colors_package not in recursive_listing(venv_directory)

    execute_venv_tool(venv_directory, "--scope=deps")
    assert app_module in recursive_listing(venv_directory)
    assert colors_package in recursive_listing(venv_directory)

    assert canonical_venv_listing == recursive_listing(venv_directory)
    venv = assert_app(venv_directory)
    assert canonical_venv_hash == CacheHelper.dir_hash(venv.site_packages_dir)


def test_non_hermetic_issue_2004(
    tmpdir,  # type: Any
    pex_bdist,  # type: str
):
    # type: (...) -> None

    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "check_hermetic.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import os
                import sys

                def check():
                    if os.environ["PYTHONPATH"] in sys.path:
                        print("Not hermetic")
                    else:
                        print("Hermetic")
                """
            )
        )
    with safe_open(os.path.join(src, "setup.cfg"), "w") as fp:
        fp.write(
            dedent(
                """\
                [metadata]
                name = hermeticity-checker
                version = 0.0.1

                [options]
                py_modules =
                    check_hermetic

                [options.entry_points]
                console_scripts =
                    check-hermetic = check_hermetic:check
                """
            )
        )
    with safe_open(os.path.join(src, "setup.py"), "w") as fp:
        fp.write("from setuptools import setup; setup()")

    pex_root = os.path.join(str(tmpdir), "pex_root")
    check_pex = os.path.join(str(tmpdir), "check.pex")
    run_pex_command(
        args=[
            pex_bdist,
            src,
            "-o",
            check_pex,
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
        ]
    ).assert_success()

    hermetic_venv = os.path.join(str(tmpdir), "hermetic")
    non_hermetic_venv = os.path.join(str(tmpdir), "non-hermetic")

    run_pex_tools(check_pex, "venv", hermetic_venv).assert_success()
    run_pex_tools(check_pex, "venv", "--non-hermetic-scripts", non_hermetic_venv).assert_success()

    hermetic_check = subprocess.check_output(
        args=[os.path.join(hermetic_venv, "bin", "check-hermetic")],
        env=make_env(PYTHONPATH=src),
    )
    non_hermetic_check = subprocess.check_output(
        args=[os.path.join(non_hermetic_venv, "bin", "check-hermetic")],
        env=make_env(PYTHONPATH=src),
    )

    assert "Hermetic" in str(hermetic_check)
    assert "Not hermetic" in str(non_hermetic_check)



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import multiprocessing
import os

from pex.common import temporary_dir
from pex.interpreter import PythonInterpreter
from pex.testing import (
    PY39,
    PY310,
    IntegResults,
    ensure_python_interpreter,
    make_env,
    run_pex_command,
    run_simple_pex,
)
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List


def test_resolve_local_platform():
    # type: () -> None
    python39 = ensure_python_interpreter(PY39)
    python310 = ensure_python_interpreter(PY310)
    pex_python_path = os.pathsep.join((python39, python310))

    def create_platform_pex(args):
        # type: (List[str]) -> IntegResults
        return run_pex_command(
            args=["--platform", str(PythonInterpreter.from_binary(python310).platform)] + args,
            python=python39,
            env=make_env(PEX_PYTHON_PATH=pex_python_path),
        )

    with temporary_dir() as td:
        pex_file = os.path.join(td, "pex_file")

        # N.B.: We use psutil since only an sdist is available for linux and osx and the
        # distribution has no dependencies.
        args = ["psutil==5.7.0", "-o", pex_file]

        # By default, no --platforms are resolved and so distributions must be available in binary
        # form.
        results = create_platform_pex(args)
        results.assert_failure()

        # If --platform resolution is enabled however, we should be able to find a corresponding
        # local interpreter to perform a full-featured resolve with.
        results = create_platform_pex(["--resolve-local-platforms"] + args)
        results.assert_success()

        output, returncode = run_simple_pex(
            pex=pex_file,
            args=("-c", "import psutil; print(psutil.cpu_count())"),
            interpreter=PythonInterpreter.from_binary(python310),
        )
        assert 0 == returncode
        assert int(output.strip()) >= multiprocessing.cpu_count()



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from argparse import ArgumentParser

import pytest


@pytest.fixture
def parser():
    # type: () -> ArgumentParser
    return ArgumentParser()



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import itertools
import os
import site
import sys
from collections import OrderedDict, defaultdict

from pex import dist_metadata, pex_warnings, targets
from pex.common import pluralize
from pex.dist_metadata import Distribution, Requirement
from pex.fingerprinted_distribution import FingerprintedDistribution
from pex.inherit_path import InheritPath
from pex.layout import maybe_install
from pex.orderedset import OrderedSet
from pex.pep_425 import CompatibilityTags, TagRank
from pex.pep_503 import ProjectName
from pex.pex_info import PexInfo
from pex.targets import Target
from pex.third_party.packaging import specifiers
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import (
        DefaultDict,
        Dict,
        FrozenSet,
        Iterable,
        Iterator,
        List,
        MutableMapping,
        Optional,
        Tuple,
        Union,
    )

    import attr  # vendor:skip
else:
    from pex.third_party import attr


def _import_pkg_resources():
    try:
        import pkg_resources  # vendor:skip

        return pkg_resources, False
    except ImportError:
        from pex import third_party

        third_party.install(expose=["setuptools"])
        import pkg_resources  # vendor:skip

        return pkg_resources, True


@attr.s(frozen=True)
class _RankedDistribution(object):
    # N.B.: A distribution implements rich comparison with the leading component being the
    # `parsed_version`; as such, a _RankedDistribution sorts as a whole 1st by `rank` (which is a
    # rank of the distribution's tags specificity for the target interpreter), then by version and
    # finally by redundant components of distribution metadata we never get to since they are
    # encoded in the tag specificity rank value.

    # The attr project type stub file simply misses this.
    _fd_cmp = attr.cmp_using(  # type: ignore[attr-defined]
        eq=FingerprintedDistribution.__eq__,
        # Since we want to rank higher versions higher (earlier) we need to reverse the natural
        # ordering of Version in Distribution which is least to greatest.
        lt=FingerprintedDistribution.__ge__,
    )

    @classmethod
    def highest_rank(cls, fingerprinted_distribution):
        # type: (FingerprintedDistribution) -> _RankedDistribution
        return cls(
            rank=TagRank.highest_natural().higher(),
            fingerprinted_distribution=fingerprinted_distribution,
        )

    rank = attr.ib()  # type: TagRank
    fingerprinted_distribution = attr.ib(
        eq=_fd_cmp, order=_fd_cmp
    )  # type: FingerprintedDistribution

    @property
    def distribution(self):
        # type: () -> Distribution
        return self.fingerprinted_distribution.distribution

    @property
    def fingerprint(self):
        # type: () -> str
        return self.fingerprinted_distribution.fingerprint

    def satisfies(self, requirement):
        # type: (Requirement) -> bool
        return self.distribution in requirement


@attr.s(frozen=True)
class _UnrankedDistribution(object):
    fingerprinted_distribution = attr.ib()  # type: FingerprintedDistribution

    @property
    def dist(self):
        # type: () -> Distribution
        return self.fingerprinted_distribution.distribution

    def render_message(self, target):
        # type: (Target) -> str
        return "The distribution {dist} cannot be used by {target}.".format(
            dist=self.dist, target=target
        )


@attr.s(frozen=True)
class _InvalidWheelName(_UnrankedDistribution):
    filename = attr.ib()  # type: str

    def render_message(self, _target):
        # type: (Target) -> str
        return (
            "The filename of {dist} is not a valid wheel file name that can be parsed for "
            "tags.".format(dist=self.dist)
        )


@attr.s(frozen=True)
class _TagMismatch(_UnrankedDistribution):
    wheel_tags = attr.ib()  # type: CompatibilityTags

    def render_message(self, target):
        # type: (Target) -> str
        return (
            "The wheel tags for {dist} are {wheel_tags} which do not match the supported tags of "
            "{target}:\n{tag}\n... {count} more ...".format(
                dist=self.dist,
                wheel_tags=", ".join(map(str, self.wheel_tags)),
                target=target,
                tag=target.supported_tags[0],
                count=len(target.supported_tags) - 1,
            )
        )


@attr.s(frozen=True)
class _PythonRequiresMismatch(_UnrankedDistribution):
    python_requires = attr.ib()  # type: specifiers.SpecifierSet

    def render_message(self, target):
        # type: (Target) -> str
        return (
            "The distribution has a python requirement of {python_requires} which does not match "
            "the python version of {python_version} for {target}.".format(
                python_requires=self.python_requires,
                python_version=target.python_version_str,
                target=target,
            )
        )


@attr.s(frozen=True)
class _QualifiedRequirement(object):
    requirement = attr.ib()  # type: Requirement
    required = attr.ib(default=True)  # type: bool


@attr.s(frozen=True)
class _DistributionNotFound(object):
    requirement = attr.ib()  # type: Requirement
    required_by = attr.ib(default=None)  # type: Optional[Distribution]


if TYPE_CHECKING:
    QualifiedRequirementOrNotFound = Union[_QualifiedRequirement, _DistributionNotFound]


class ResolveError(Exception):
    """Indicates an error resolving requirements from within a PEX."""


@attr.s(frozen=True)
class _RequirementKey(ProjectName):
    @classmethod
    def create(cls, requirement):
        # type: (Requirement) -> _RequirementKey
        return cls(requirement.name, frozenset(requirement.extras))

    extras = attr.ib()  # type: FrozenSet[str]

    def satisfied_keys(self):
        # type: () -> Iterator[_RequirementKey]

        # If we resolve a requirement with extras then we've satisfied resolves for the powerset of
        # the extras.
        # For example, if we resolve `cake[birthday,wedding]` then we satisfy resolves for:
        # `cake[]`
        # `cake[birthday]`
        # `cake[wedding]`
        # `cake[birthday,wedding]`
        items = list(self.extras)
        for size in range(len(items) + 1):
            for combination_of_size in itertools.combinations(items, size):
                yield _RequirementKey(self.raw, frozenset(combination_of_size))


class PEXEnvironment(object):
    _CACHE = {}  # type: Dict[Tuple[str, str, Target], PEXEnvironment]

    @classmethod
    def mount(
        cls,
        pex,  # type: str
        pex_info=None,  # type: Optional[PexInfo]
        target=None,  # type: Optional[Target]
    ):
        # type: (...) -> PEXEnvironment
        pex_file = os.path.realpath(pex)
        if not pex_info:
            pex_info = PexInfo.from_pex(pex_file)
            pex_info.update(PexInfo.from_env())
        pex_hash = pex_info.pex_hash
        if pex_hash is None:
            raise AssertionError(
                "There was no pex_hash stored in {} for {}.".format(PexInfo.PATH, pex)
            )
        target = target or targets.current()
        key = (pex_file, pex_hash, target)
        mounted = cls._CACHE.get(key)
        if mounted is None:
            pex_root = pex_info.pex_root
            pex = maybe_install(pex=pex, pex_root=pex_root, pex_hash=pex_hash) or pex
            mounted = cls(pex=pex, pex_info=pex_info, target=target)
            cls._CACHE[key] = mounted
        return mounted

    def __init__(
        self,
        pex,  # type: str
        pex_info=None,  # type: Optional[PexInfo]
        target=None,  # type: Optional[Target]
    ):
        # type: (...) -> None
        self._pex = os.path.realpath(pex)
        self._pex_info = pex_info or PexInfo.from_pex(pex)
        self._target = target or targets.current()

        self._available_ranked_dists_by_project_name = defaultdict(
            list
        )  # type: DefaultDict[ProjectName, List[_RankedDistribution]]
        self._unavailable_dists_by_project_name = defaultdict(
            list
        )  # type: DefaultDict[ProjectName, List[_UnrankedDistribution]]
        self._resolved_dists = None  # type: Optional[Iterable[Distribution]]
        self._activated_dists = None  # type: Optional[Iterable[Distribution]]

    @property
    def path(self):
        # type: () -> str
        return self._pex

    def iter_distributions(self):
        # type: () -> Iterator[FingerprintedDistribution]
        internal_cache = os.path.join(self._pex, self._pex_info.internal_cache)
        with TRACER.timed("Searching dependency cache: %s" % internal_cache, V=2):
            for distribution_name, fingerprint in self._pex_info.distributions.items():
                dist_path = os.path.join(internal_cache, distribution_name)
                yield FingerprintedDistribution(
                    distribution=Distribution.load(dist_path),
                    fingerprint=fingerprint,
                )

    def _update_candidate_distributions(self, distribution_iter):
        # type: (Iterable[FingerprintedDistribution]) -> None
        for fingerprinted_dist in distribution_iter:
            ranked_dist = self._can_add(fingerprinted_dist)
            project_name = fingerprinted_dist.project_name
            if isinstance(ranked_dist, _RankedDistribution):
                with TRACER.timed("Adding %s" % fingerprinted_dist.distribution, V=2):
                    self._available_ranked_dists_by_project_name[project_name].append(ranked_dist)
            else:
                self._unavailable_dists_by_project_name[project_name].append(ranked_dist)

    def _can_add(self, fingerprinted_dist):
        # type: (FingerprintedDistribution) -> Union[_RankedDistribution, _UnrankedDistribution]
        filename, ext = os.path.splitext(os.path.basename(fingerprinted_dist.location))
        if ext.lower() != ".whl":
            # This supports resolving pex's own vendored distributions which are vendored in a
            # directory with the project name (`pip/` for pip) and not the corresponding wheel name
            # (`pip-19.3.1-py2.py3-none-any.whl/` for pip). Pex only vendors universal wheels for
            # all platforms it supports at buildtime and runtime so this is always safe.
            return _RankedDistribution.highest_rank(fingerprinted_dist)

        try:
            wheel_tags = CompatibilityTags.from_wheel(fingerprinted_dist.location)
        except ValueError:
            return _InvalidWheelName(fingerprinted_dist, filename)

        # There will be multiple parsed tags for compressed tag sets. Ensure we grab the parsed tag
        # with highest rank from that expanded set.
        best_match = self._target.supported_tags.best_match(wheel_tags)
        if best_match is None:
            return _TagMismatch(fingerprinted_dist, wheel_tags)

        python_requires = dist_metadata.requires_python(fingerprinted_dist.distribution)
        if python_requires and not self._target.requires_python_applies(
            python_requires, source=fingerprinted_dist.distribution.as_requirement()
        ):
            return _PythonRequiresMismatch(fingerprinted_dist, python_requires)

        return _RankedDistribution(best_match.rank, fingerprinted_dist)

    def activate(self):
        # type: () -> Iterable[Distribution]
        if self._activated_dists is None:
            with TRACER.timed("Activating PEX virtual environment from %s" % self._pex):
                self._activated_dists = self._activate()
        return self._activated_dists

    def _evaluate_marker(
        self,
        requirement,  # type: Requirement
        extras=(),  # type: Iterable[str]
    ):
        # type: (...) -> bool
        applies = self._target.requirement_applies(requirement, extras=extras)
        if not applies:
            TRACER.log(
                "Skipping activation of `{}` due to environment marker de-selection".format(
                    requirement
                ),
                V=3,
            )
        return applies

    def _resolve_requirement(
        self,
        requirement,  # type: Requirement
        resolved_dists_by_key,  # type: MutableMapping[_RequirementKey, FingerprintedDistribution]
        required,  # type: bool
        required_by=None,  # type: Optional[Distribution]
    ):
        # type: (...) -> Iterator[_DistributionNotFound]
        requirement_key = _RequirementKey.create(requirement)
        if requirement_key in resolved_dists_by_key:
            return

        available_distributions = [
            ranked_dist
            for ranked_dist in self._available_ranked_dists_by_project_name[
                requirement.project_name
            ]
            if ranked_dist.satisfies(requirement)
        ]
        if not available_distributions:
            if required:
                yield _DistributionNotFound(requirement, required_by=required_by)
            return

        resolved_distribution = sorted(available_distributions)[0].fingerprinted_distribution
        if len(available_distributions) > 1:
            TRACER.log(
                "Resolved {req} to {dist} and discarded {discarded}.".format(
                    req=requirement,
                    dist=resolved_distribution.distribution,
                    discarded=", ".join(
                        str(ranked_dist.distribution) for ranked_dist in available_distributions[1:]
                    ),
                ),
                V=9,
            )

        resolved_dists_by_key.update(
            (key, resolved_distribution) for key in requirement_key.satisfied_keys()
        )

        for dep_requirement in dist_metadata.requires_dists(resolved_distribution.distribution):
            # A note regarding extras and why they're passed down one level (we don't pass / use
            # dep_requirement.extras for example):
            #
            # Say we're resolving the `requirement` 'requests[security]==2.25.1'. That means
            # `resolved_distribution` is the requests distribution. It will have metadata that
            # looks like so:
            #
            # $ grep Requires-Dist requests-2.25.1.dist-info/METADATA | grep security -C1
            # Requires-Dist: certifi (>=2017.4.17)
            # Requires-Dist: pyOpenSSL (>=0.14) ; extra == 'security'
            # Requires-Dist: cryptography (>=1.3.4) ; extra == 'security'
            # Requires-Dist: PySocks (!=1.5.7,>=1.5.6) ; extra == 'socks'
            #
            # We want to recurse and resolve all standard requests requirements but also those that
            # are part of the 'security' extra. In order to resolve the latter we need to include
            # the 'security' extra environment marker.
            required = self._evaluate_marker(dep_requirement, extras=requirement.extras)
            if not required:
                continue

            for not_found in self._resolve_requirement(
                dep_requirement,
                resolved_dists_by_key,
                required,
                required_by=resolved_distribution.distribution,
            ):
                yield not_found

    def _root_requirements_iter(self, reqs):
        # type: (Iterable[Requirement]) -> Iterator[QualifiedRequirementOrNotFound]

        # We want to pick one requirement for each key (required project) to then resolve
        # recursively.

        # First, the selected requirement clearly needs to be applicable (its environment markers
        # must apply to our interpreter). For example, for a Python 3.6 interpreter this would
        # select just "isort==5.6.4; python_version>='3.6'" from the input set:
        # {
        #   "isort==4.3.21; python_version<'3.6'",
        #   "setuptools==44.1.1; python_version<'3.6'",
        #   "isort==5.6.4; python_version>='3.6'",
        # }
        qualified_reqs_by_project_name = (
            OrderedDict()
        )  # type: OrderedDict[ProjectName, List[_QualifiedRequirement]]
        for req in reqs:
            required = self._evaluate_marker(req)
            if not required:
                continue
            project_name = req.project_name
            requirements = qualified_reqs_by_project_name.get(project_name)
            if requirements is None:
                qualified_reqs_by_project_name[project_name] = requirements = []
            requirements.append(_QualifiedRequirement(req, required=required))

        # Next, from among the remaining applicable requirements for a given project, we want to
        # select the most tailored (highest ranked) available distribution. That distribution's
        # transitive requirements will later fill in the full resolve.
        for project_name, qualified_requirements in qualified_reqs_by_project_name.items():
            ranked_dists = self._available_ranked_dists_by_project_name.get(project_name)
            if ranked_dists is None:
                # We've winnowed down reqs_by_key to just those requirements whose environment
                # markers apply; so, we should always have an available distribution.
                message = (
                    "A distribution for {project_name} could not be resolved for {target}.".format(
                        project_name=project_name, target=self._target
                    )
                )
                unavailable_dists = self._unavailable_dists_by_project_name.get(project_name)
                if unavailable_dists:
                    message += (
                        "\nFound {count} {distributions} for {project_name} that do not apply:\n"
                        "{unavailable_dists}".format(
                            count=len(unavailable_dists),
                            distributions=pluralize(unavailable_dists, "distribution"),
                            project_name=project_name,
                            unavailable_dists="\n".join(
                                "{index}.) {message}".format(
                                    index=index,
                                    message=unavailable_dist.render_message(self._target),
                                )
                                for index, unavailable_dist in enumerate(unavailable_dists, start=1)
                            ),
                        )
                    )
                raise ResolveError(message)
            candidates = [
                (ranked_dist, qualified_requirement)
                for qualified_requirement in qualified_requirements
                for ranked_dist in ranked_dists
                if ranked_dist.satisfies(qualified_requirement.requirement)
            ]
            if not candidates:
                for qualified_requirement in qualified_requirements:
                    yield _DistributionNotFound(qualified_requirement.requirement)
                continue

            ranked_dist, qualified_requirement = sorted(candidates, key=lambda tup: tup[0])[0]
            if len(candidates) > 1:
                TRACER.log(
                    "Selected {dist} via {req} and discarded {discarded}.".format(
                        req=qualified_requirement.requirement,
                        dist=ranked_dist.distribution,
                        discarded=", ".join(
                            "{dist} via {req}".format(
                                req=qualified_req.requirement, dist=ranked_dist.distribution
                            )
                            for ranked_dist, qualified_req in candidates[1:]
                        ),
                    ),
                    V=9,
                )
            yield qualified_requirement

    def resolve(self):
        # type: () -> Iterable[Distribution]
        if self._resolved_dists is None:
            all_reqs = [Requirement.parse(req) for req in self._pex_info.requirements]
            self._resolved_dists = tuple(
                fingerprinted_distribution.distribution
                for fingerprinted_distribution in self.resolve_dists(all_reqs)
            )
        return self._resolved_dists

    def resolve_dists(self, reqs):
        # type: (Iterable[Requirement]) -> Iterable[FingerprintedDistribution]

        self._update_candidate_distributions(self.iter_distributions())

        unresolved_reqs = OrderedDict()  # type: OrderedDict[Requirement, OrderedSet]

        def record_unresolved(dist_not_found):
            # type: (_DistributionNotFound) -> None
            TRACER.log("Failed to resolve a requirement: {}".format(dist_not_found.requirement))
            requirers = unresolved_reqs.get(dist_not_found.requirement)
            if requirers is None:
                requirers = OrderedSet()
                unresolved_reqs[dist_not_found.requirement] = requirers
            if dist_not_found.required_by:
                requirers.add(dist_not_found.required_by)

        resolved_dists_by_key = (
            OrderedDict()
        )  # type: OrderedDict[_RequirementKey, FingerprintedDistribution]
        for qualified_req_or_not_found in self._root_requirements_iter(reqs):
            if isinstance(qualified_req_or_not_found, _DistributionNotFound):
                record_unresolved(qualified_req_or_not_found)
                continue

            with TRACER.timed("Resolving {}".format(qualified_req_or_not_found.requirement), V=2):
                for not_found in self._resolve_requirement(
                    requirement=qualified_req_or_not_found.requirement,
                    required=qualified_req_or_not_found.required,
                    resolved_dists_by_key=resolved_dists_by_key,
                ):
                    record_unresolved(not_found)

        if unresolved_reqs:
            TRACER.log("Unresolved requirements:")
            for req in unresolved_reqs:
                TRACER.log("  - {}".format(req))

            TRACER.log("Distributions contained within this pex:")
            if not self._pex_info.distributions:
                TRACER.log("  None")
            else:
                for dist_name in self._pex_info.distributions:
                    TRACER.log("  - {}".format(dist_name))

            if not self._pex_info.ignore_errors:
                items = []
                for index, (requirement, requirers) in enumerate(unresolved_reqs.items()):
                    rendered_requirers = ""
                    if requirers:
                        rendered_requirers = "\n    Required by:" "\n      {requirers}".format(
                            requirers="\n      ".join(map(str, requirers))
                        )
                    contains = self._available_ranked_dists_by_project_name[
                        requirement.project_name
                    ]
                    if contains:
                        rendered_contains = (
                            "\n    But this pex only contains:"
                            "\n      {distributions}".format(
                                distributions="\n      ".join(
                                    os.path.basename(ranked_dist.distribution.location)
                                    for ranked_dist in contains
                                ),
                            )
                        )
                    else:
                        rendered_contains = (
                            "\n    But this pex had no {project_name!r} distributions.".format(
                                project_name=requirement.project_name
                            )
                        )
                    items.append(
                        "{index: 2d}: {requirement}"
                        "{rendered_requirers}"
                        "{rendered_contains}".format(
                            index=index + 1,
                            requirement=requirement,
                            rendered_requirers=rendered_requirers,
                            rendered_contains=rendered_contains,
                        )
                    )

                raise ResolveError(
                    "Failed to resolve requirements from PEX environment @ {pex}.\n"
                    "Needed {platform} compatible dependencies for:\n"
                    "{items}".format(
                        pex=self._pex, platform=self._target.platform.tag, items="\n".join(items)
                    )
                )

        return OrderedSet(resolved_dists_by_key.values())

    _NAMESPACE_PACKAGE_METADATA_RESOURCE = "namespace_packages.txt"

    @classmethod
    def _get_namespace_packages(cls, dist):
        if dist.has_metadata(cls._NAMESPACE_PACKAGE_METADATA_RESOURCE):
            return list(dist.get_metadata_lines(cls._NAMESPACE_PACKAGE_METADATA_RESOURCE))
        else:
            return []

    @classmethod
    def _declare_namespace_packages(cls, resolved_dists):
        # type: (Iterable[Distribution]) -> None
        namespace_packages_by_dist = OrderedDict()
        for dist in resolved_dists:
            namespace_packages = cls._get_namespace_packages(dist)
            # NB: Dists can explicitly declare empty namespace packages lists to indicate they have none.
            # We only care about dists with one or more namespace packages though; thus, the guard.
            if namespace_packages:
                namespace_packages_by_dist[dist] = namespace_packages

        if not namespace_packages_by_dist:
            return  # Nothing to do here.

        # When declaring namespace packages, we need to do so with the `setuptools` distribution that
        # will be active in the pex environment at runtime and, as such, care must be taken.
        #
        # Properly behaved distributions will declare a dependency on `setuptools`, in which case we
        # use that (non-vendored) distribution. A side-effect of importing `pkg_resources` from that
        # distribution is that a global `pkg_resources.working_set` will be populated. For various
        # `pkg_resources` distribution discovery functions to work, that global
        # `pkg_resources.working_set` must be built with the `sys.path` fully settled. Since all dists
        # in the dependency set (`resolved_dists`) have already been resolved and added to the
        # `sys.path` we're safe to proceed here.
        #
        # Other distributions (notably `twitter.common.*`) in the wild declare `setuptools`-specific
        # `namespace_packages` but do not properly declare a dependency on `setuptools` which they must
        # use to:
        # 1. Declare `namespace_packages` metadata which we just verified they have with the check
        #    above.
        # 2. Declare namespace packages at runtime via the canonical:
        #    `__import__('pkg_resources').declare_namespace(__name__)`
        #
        # For such distributions we fall back to our vendored version of `setuptools`. This is safe,
        # since we'll only introduce our shaded version when no other standard version is present and
        # even then tear it all down when we hand off from the bootstrap to user code.
        pkg_resources, vendored = _import_pkg_resources()
        if vendored:
            dists = "\n".join(
                "\n{index}. {dist} namespace packages:\n  {ns_packages}".format(
                    index=index + 1,
                    dist=dist.as_requirement(),
                    ns_packages="\n  ".join(ns_packages),
                )
                for index, (dist, ns_packages) in enumerate(namespace_packages_by_dist.items())
            )
            pex_warnings.warn(
                "The `pkg_resources` package was loaded from a pex vendored version when "
                "declaring namespace packages defined by:\n{dists}\n\nThese distributions "
                "should fix their `install_requires` to include `setuptools`".format(dists=dists)
            )

        for pkg in itertools.chain(*namespace_packages_by_dist.values()):
            if pkg in sys.modules:
                pkg_resources.declare_namespace(pkg)

    def _activate(self):
        # type: () -> Iterable[Distribution]

        if not any(self._pex == os.path.realpath(path) for path in sys.path):
            TRACER.log("Adding pex environment to the head of sys.path: {}".format(self._pex))
            sys.path.insert(0, self._pex)

        resolved = self.resolve()
        for dist in resolved:
            # N.B.: Since there can be more than one PEXEnvironment on the PEX_PATH we take care to
            # avoid re-installing duplicate distributions we have in common with them.
            if dist.location in sys.path:
                continue
            with TRACER.timed("Activating %s" % dist, V=2):
                if self._pex_info.inherit_path == InheritPath.FALLBACK:
                    # Prepend location to sys.path.
                    #
                    # This ensures that bundled versions of libraries will be used before system-installed
                    # versions, in case something is installed in both, helping to favor hermeticity in
                    # the case of non-hermetic PEX files (i.e. those with inherit_path=True).
                    #
                    # If the path is not already in sys.path, site.addsitedir will append (not prepend)
                    # the path to sys.path. But if the path is already in sys.path, site.addsitedir will
                    # leave sys.path unmodified, but will do everything else it would do. This is not part
                    # of its advertised contract (which is very vague), but has been verified to be the
                    # case by inspecting its source for both cpython 2.7 and cpython 3.7.
                    sys.path.insert(0, dist.location)
                else:
                    sys.path.append(dist.location)

                with TRACER.timed("Adding sitedir", V=2):
                    site.addsitedir(dist.location)
        return resolved



import os
from glob import glob
from distutils.util import convert_path
from distutils.command import sdist

if "__PEX_UNVENDORED__" in __import__("os").environ:
  from setuptools.extern.six.moves import filter  # vendor:skip
else:
  from pex.third_party.setuptools.extern.six.moves import filter



class sdist_add_defaults:
    """
    Mix-in providing forward-compatibility for functionality as found in
    distutils on Python 3.7.

    Do not edit the code in this class except to update functionality
    as implemented in distutils. Instead, override in the subclass.
    """

    def add_defaults(self):
        """Add all the default files to self.filelist:
          - README or README.txt
          - setup.py
          - test/test*.py
          - all pure Python modules mentioned in setup script
          - all files pointed by package_data (build_py)
          - all files defined in data_files.
          - all files defined as scripts.
          - all C sources listed as part of extensions or C libraries
            in the setup script (doesn't catch C headers!)
        Warns if (README or README.txt) or setup.py are missing; everything
        else is optional.
        """
        self._add_defaults_standards()
        self._add_defaults_optional()
        self._add_defaults_python()
        self._add_defaults_data_files()
        self._add_defaults_ext()
        self._add_defaults_c_libs()
        self._add_defaults_scripts()

    @staticmethod
    def _cs_path_exists(fspath):
        """
        Case-sensitive path existence check

        >>> sdist_add_defaults._cs_path_exists(__file__)
        True
        >>> sdist_add_defaults._cs_path_exists(__file__.upper())
        False
        """
        if not os.path.exists(fspath):
            return False
        # make absolute so we always have a directory
        abspath = os.path.abspath(fspath)
        directory, filename = os.path.split(abspath)
        return filename in os.listdir(directory)

    def _add_defaults_standards(self):
        standards = [self.READMES, self.distribution.script_name]
        for fn in standards:
            if isinstance(fn, tuple):
                alts = fn
                got_it = False
                for fn in alts:
                    if self._cs_path_exists(fn):
                        got_it = True
                        self.filelist.append(fn)
                        break

                if not got_it:
                    self.warn("standard file not found: should have one of " +
                              ', '.join(alts))
            else:
                if self._cs_path_exists(fn):
                    self.filelist.append(fn)
                else:
                    self.warn("standard file '%s' not found" % fn)

    def _add_defaults_optional(self):
        optional = ['test/test*.py', 'setup.cfg']
        for pattern in optional:
            files = filter(os.path.isfile, glob(pattern))
            self.filelist.extend(files)

    def _add_defaults_python(self):
        # build_py is used to get:
        #  - python modules
        #  - files defined in package_data
        build_py = self.get_finalized_command('build_py')

        # getting python files
        if self.distribution.has_pure_modules():
            self.filelist.extend(build_py.get_source_files())

        # getting package_data files
        # (computed in build_py.data_files by build_py.finalize_options)
        for pkg, src_dir, build_dir, filenames in build_py.data_files:
            for filename in filenames:
                self.filelist.append(os.path.join(src_dir, filename))

    def _add_defaults_data_files(self):
        # getting distribution.data_files
        if self.distribution.has_data_files():
            for item in self.distribution.data_files:
                if isinstance(item, str):
                    # plain file
                    item = convert_path(item)
                    if os.path.isfile(item):
                        self.filelist.append(item)
                else:
                    # a (dirname, filenames) tuple
                    dirname, filenames = item
                    for f in filenames:
                        f = convert_path(f)
                        if os.path.isfile(f):
                            self.filelist.append(f)

    def _add_defaults_ext(self):
        if self.distribution.has_ext_modules():
            build_ext = self.get_finalized_command('build_ext')
            self.filelist.extend(build_ext.get_source_files())

    def _add_defaults_c_libs(self):
        if self.distribution.has_c_libraries():
            build_clib = self.get_finalized_command('build_clib')
            self.filelist.extend(build_clib.get_source_files())

    def _add_defaults_scripts(self):
        if self.distribution.has_scripts():
            build_scripts = self.get_finalized_command('build_scripts')
            self.filelist.extend(build_scripts.get_source_files())


if hasattr(sdist.sdist, '_add_defaults_standards'):
    # disable the functionality already available upstream
    class sdist_add_defaults:
        pass



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
import sys

from pex.compatibility import PY3
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.resolve.locked_resolve import LocalProjectArtifact
from pex.resolve.lockfile import json_codec
from pex.resolve.resolved_requirement import Pin
from pex.testing import PY38, ensure_python_interpreter, make_env
from pex.typing import TYPE_CHECKING
from pex.version import __version__

if TYPE_CHECKING:
    from typing import Any


def test_pep_518_venv_pex_env_scrubbing(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    # N.B.: The package script requires Python 3.
    python = sys.executable if PY3 else ensure_python_interpreter(PY38)

    package_script = os.path.join(pex_project_dir, "scripts", "package.py")
    pex_pex = os.path.join(str(tmpdir), "pex")
    subprocess.check_call(args=[python, package_script, "--local", "--pex-output-file", pex_pex])

    lock = os.path.join(str(tmpdir), "lock.json")
    subprocess.check_call(
        args=[
            # Although the package script requires Python 3 to create the Pex PEX, we should be
            # able to execute the Pex PEX with any interpreter.
            sys.executable,
            pex_pex,
            "lock",
            "create",
            pex_project_dir,
            "-o",
            lock,
            "--indent",
            "2",
        ],
        env=make_env(PEX_SCRIPT="pex3"),
    )

    lockfile = json_codec.load(lock)
    assert 1 == len(lockfile.locked_resolves)

    locked_resolve = lockfile.locked_resolves[0]
    assert 1 == len(locked_resolve.locked_requirements)

    locked_requirement = locked_resolve.locked_requirements[0]
    assert Pin(ProjectName("pex"), Version(__version__)) == locked_requirement.pin
    assert isinstance(locked_requirement.artifact, LocalProjectArtifact)
    assert pex_project_dir == locked_requirement.artifact.directory
    assert not locked_requirement.additional_artifacts



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

# Due to the PEX_ properties, disable checkstyle.
# checkstyle: noqa

from __future__ import absolute_import

import hashlib
import json
import os
import re
import sys
from contextlib import contextmanager
from textwrap import dedent

from pex import pex_warnings
from pex.common import can_write_dir, die, safe_mkdtemp
from pex.inherit_path import InheritPath
from pex.orderedset import OrderedSet
from pex.typing import TYPE_CHECKING, Generic, overload
from pex.venv.bin_path import BinPath

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Iterator, Mapping, Optional, Tuple, Type, TypeVar, Union

    _O = TypeVar("_O")
    _P = TypeVar("_P")

    # N.B.: This is an expensive import and we only need it for type checking.
    from pex.interpreter import PythonInterpreter


class NoValueError(Exception):
    """Indicates a property has no value set.

    When raised from a method decorated with `@defaulted_property(default)` indicates the default
    value should be used.
    """


class DefaultedProperty(Generic["_O", "_P"]):
    """Represents a property with a default value.

    To determine the value of the property without the default value applied, access it through the
    class and call the `strip_default` method, passing the instance in question.
    """

    def __init__(
        self,
        func,  # type: Callable[[_O], _P]
        default,  # type: _P
    ):
        # type: (...) -> None
        self._func = func  # type: Callable[[_O], _P]
        self._default = default  # type: _P
        self._validator = None  # type: Optional[Callable[[_O, _P], _P]]

    @overload
    def __get__(
        self,
        instance,  # type: None
        owner_class=None,  # type: Optional[Type[_O]]
    ):
        # type: (...) -> DefaultedProperty[_O, _P]
        pass

    @overload
    def __get__(
        self,
        instance,  # type: _O
        owner_class=None,  # type: Optional[Type[_O]]
    ):
        # type: (...) -> _P
        pass

    def __get__(
        self,
        instance,  # type: Optional[_O]
        owner_class=None,  # type: Optional[Type[_O]]
    ):
        # type: (...) -> Union[DefaultedProperty[_O, _P], _P]
        if instance is None:  # The descriptor was accessed from the class.
            return self
        try:
            return self._validate(instance, self._func(instance))
        except NoValueError:
            return self._validate(instance, self._default)

    def strip_default(self, instance):
        # type: (_O) -> Optional[_P]
        """Return the value of this property without the default value applied.

        If the property is not set `None` will be returned and if there is an associated @validator
        that validator will be skipped.

        :param instance: The instance to check for the non-defaulted property value.
        :return: The property value or `None` if not set.
        """
        try:
            return self._validate(instance, self._func(instance))
        except NoValueError:
            return None

    def value_or(
        self,
        instance,  # type: _O
        fallback,  # type: _P
    ):
        # type: (...) -> _P
        """Return the value of this property without the default value applied or else the fallback.

        If the property is not set, `fallback` will be validated and returned.

        :param instance: The instance to check for the non-defaulted property value.
        :return: The property value or `fallback` if not set.
        """
        try:
            value = self._func(instance)
        except NoValueError:
            value = fallback
        return self._validate(instance, value)

    def validator(self, func):
        # type: (Callable[[_O, _P], _P]) -> Callable[[_O, _P], _P]
        """Associate a validation function with this defaulted property.

        The function will be used to validate this property's final computed value.

        :param func: The validation function to associate with this property descriptor.
        """
        self._validator = func
        return func

    def _validate(self, instance, value):
        # type: (_O, _P) -> _P
        if self._validator is None:
            return value
        return self._validator(instance, value)


def defaulted_property(
    default,  # type: _P
):
    # type: (...) -> Callable[[Callable[[_O], _P]], DefaultedProperty[_O, _P]]
    """Creates a `@property` with a `default` value.

    Accessors decorated with this function should raise `NoValueError` to indicate the `default`
    value should be used.
    """

    def wrapped(func):
        # type: (Callable[[_O], _P]) -> DefaultedProperty[_O, _P]
        return DefaultedProperty(func, default)

    return wrapped


class Variables(object):
    """Environment variables supported by the PEX runtime."""

    @classmethod
    def process_pydoc(cls, pydoc):
        # type: (Optional[str]) -> Tuple[str, str]
        if pydoc is None:
            return "Unknown", "Unknown"
        pydoc_lines = pydoc.splitlines()
        variable_type = pydoc_lines[0]
        variable_text = " ".join(filter(None, (line.strip() for line in pydoc_lines[2:])))
        return variable_type, variable_text

    @classmethod
    def iter_help(cls):
        # type: () -> Iterator[Tuple[str, str, str]]
        for variable_name, value in sorted(cls.__dict__.items()):
            if not variable_name.startswith("PEX_"):
                continue
            value = value._func if isinstance(value, DefaultedProperty) else value
            variable_type, variable_text = cls.process_pydoc(getattr(value, "__doc__"))
            yield variable_name, variable_type, variable_text

    @classmethod
    def from_rc(cls, rc=None):
        # type: (Optional[str]) -> Dict[str, str]
        """Read pex runtime configuration variables from a pexrc file.

        :param rc: an absolute path to a pexrc file.
        :return: A dict of key value pairs found in processed pexrc files.
        """
        ret_vars = {}  # type: Dict[str, str]
        rc_locations = [
            os.path.join(os.sep, "etc", "pexrc"),
            os.path.join("~", ".pexrc"),
            os.path.join(os.path.dirname(sys.argv[0]), ".pexrc"),
        ]
        if rc:
            rc_locations.append(rc)
        for filename in rc_locations:
            try:
                with open(os.path.expanduser(filename)) as fh:
                    rc_items = map(cls._get_kv, fh)
                    ret_vars.update(dict(filter(None, rc_items)))
            except IOError:
                continue
        return ret_vars

    @classmethod
    def _get_kv(cls, variable):
        kv = variable.strip().split("=")
        if len(list(filter(None, kv))) == 2:
            return kv

    @staticmethod
    def _maybe_get_bool_var(
        name,  # type: str
        env,  # type: Mapping[str, Optional[str]]
    ):
        # type: (...) -> Optional[bool]
        value = env.get(name, None)
        if value is None:
            return None
        if value.lower() in ("0", "false"):
            return False
        if value.lower() in ("1", "true"):
            return True
        raise ValueError(
            "Invalid bool value for {name}, must be 0/1/false/true, got {value!r}".format(
                name=name, value=value
            )
        )

    def __init__(self, environ=None, rc=None):
        # type: (Optional[Mapping[str, str]], Optional[str]) -> None
        env = environ if environ is not None else os.environ
        if self._maybe_get_bool_var("PEX_DISABLE_VARIABLES", env) is True:
            self._environ = {
                key: value
                for key, value in env.items()
                if key == "PEX_DISABLE_VARIABLES" or not key.startswith("PEX_")
            }
        else:
            self._environ = dict(env)
            if not self.PEX_IGNORE_RCFILES:
                rc_values = self.from_rc(rc).copy()
                rc_values.update(self._environ)
                self._environ = rc_values

        if "PEX_ALWAYS_CACHE" in self._environ:
            pex_warnings.warn(
                "The `PEX_ALWAYS_CACHE` env var is deprecated. This env var is no longer read; all "
                "internally cached distributions in a PEX are always installed into the local Pex "
                "dependency cache."
            )
        if "PEX_FORCE_LOCAL" in self._environ:
            pex_warnings.warn(
                "The `PEX_FORCE_LOCAL` env var is deprecated. This env var is no longer read since "
                "user code is now always unzipped before execution."
            )
        if "PEX_UNZIP" in self._environ:
            pex_warnings.warn(
                "The `PEX_UNZIP` env var is deprecated. This env var is no longer read since "
                "unzipping PEX zip files before execution is now the default."
            )
        if "PEX_TEARDOWN_VERBOSE" in self._environ:
            pex_warnings.warn(
                "The `PEX_TEARDOWN_VERBOSE` env var is deprecated. This env var is no longer read "
                "since PEX teardown has been removed in favor of the natural teardown environment "
                "provided by the Python runtime."
            )

    def copy(self):
        # type: () -> Dict[str, str]
        return self._environ.copy()

    def _maybe_get_string(self, variable):
        # type: (str) -> Optional[str]
        return self._environ.get(variable)

    def _get_string(self, variable):
        # type: (str) -> str
        value = self._maybe_get_string(variable)
        if value is None:
            raise NoValueError(variable)
        return value

    def _maybe_get_bool(self, variable):
        # type: (str) -> Optional[bool]
        return self._maybe_get_bool_var(variable, self._environ)

    def _get_bool(self, variable):
        # type: (str) -> bool
        value = self._maybe_get_bool(variable)
        if value is None:
            raise NoValueError(variable)
        return value

    def _maybe_get_path(self, variable):
        # type: (str) -> Optional[str]
        value = self._maybe_get_string(variable)
        if value is None:
            return None
        return os.path.realpath(os.path.expanduser(value))

    def _get_path(self, variable):
        # type: (str) -> str
        value = self._maybe_get_path(variable)
        if value is None:
            raise NoValueError(variable)
        return value

    def _get_int(self, variable):
        # type: (str) -> int
        value = self._get_string(variable)
        try:
            return int(value)
        except ValueError:
            die(
                "Invalid value for %s, must be an integer, got %r"
                % (variable, self._environ[variable])
            )

    def _maybe_get_path_tuple(
        self,
        variable,  # type: str
        empty_string_is_cwd=True,  # type: bool
    ):
        # type: (...) -> Optional[Tuple[str, ...]]
        value = self._maybe_get_string(variable)
        if value is None:
            return None
        if not value and not empty_string_is_cwd:
            return None
        return tuple(
            OrderedSet(os.path.normpath(os.path.expanduser(p)) for p in value.split(os.pathsep))
        )

    def strip(self):
        # type: () -> Variables
        stripped_environ = {
            k: v for k, v in self.copy().items() if not k.startswith(("PEX_", "__PEX_"))
        }
        return Variables(environ=stripped_environ)

    @contextmanager
    def patch(self, **kw):
        # type: (**Optional[str]) -> Iterator[Dict[str, str]]
        """Update the environment for the duration of a context.

        Any environment variable with a value of `None` will be removed from the environment if
        present. The rest will be added to the environment or else updated if already present in
        the environment.
        """
        disable_env = self._maybe_get_bool_var("PEX_DISABLE_VARIABLES", kw)

        old_environ = self._environ
        self._environ = self._environ.copy()
        if disable_env:
            for k in list(self._environ):
                if k != "PEX_DISABLE_VARIABLES" and k.startswith("PEX_"):
                    self._environ.pop(k)
        for k, v in kw.items():
            if v is None:
                self._environ.pop(k, None)
            elif disable_env and k != "PEX_DISABLE_VARIABLES" and k.startswith("PEX_"):
                self._environ.pop(k, None)
            else:
                self._environ[k] = v

        yield self._environ
        self._environ = old_environ

    @defaulted_property(default=False)
    def PEX_DISABLE_VARIABLES(self):
        # type: () -> bool
        """Boolean.

        Disable reading of all PEX_* variables (except this one) from all sources. Both PEX_*
        environment variables and PEX_* variables sources from pexrc files will be ignored.

        This can be used to lock down the function of a PEX.
        """
        return self._get_bool("PEX_DISABLE_VARIABLES")

    @property
    def PEX(self):
        # type: () -> Optional[str]
        """String.

        The absolute path of the PEX that is executing. This will either be the path of the PEX
        zip for zipapps or else the path of the PEX loose directory for packed or loose PEXes.

        N.B.: The path is not the path of the underlying unzipped PEX or else the PEXes installed
        venv.
        """
        return self._maybe_get_path("PEX")

    @defaulted_property(default=False)
    def PEX_ALWAYS_CACHE(self):
        # type: () -> bool
        """Boolean.

        Deprecated: This env var is no longer used; all internally cached distributions in a PEX
        are always installed into the local Pex dependency cache.
        """
        return self._get_bool("PEX_ALWAYS_CACHE")

    @defaulted_property(default=False)
    def PEX_COVERAGE(self):
        # type: () -> bool
        """Boolean.

        Enable coverage reporting for this PEX file.  This requires that the "coverage" module is
        available in the PEX environment.

        Default: false.
        """
        return self._get_bool("PEX_COVERAGE")

    @property
    def PEX_COVERAGE_FILENAME(self):
        # type: () -> Optional[str]
        """Filename.

        Write the coverage data to the specified filename.  If PEX_COVERAGE_FILENAME is not
        specified but PEX_COVERAGE is, coverage information will be printed to stdout and not saved.
        """
        return self._maybe_get_path("PEX_COVERAGE_FILENAME")

    @defaulted_property(default=False)
    def PEX_FORCE_LOCAL(self):
        # type: () -> bool
        """Boolean.

        Deprecated: This env var is no longer used since user code is now always unzipped before
        execution.
        """
        return self._get_bool("PEX_FORCE_LOCAL")

    @defaulted_property(default=False)
    def PEX_UNZIP(self):
        # type: () -> bool
        """Boolean.

        Deprecated: This env var is no longer used since unzipping PEX zip files before execution
        is now the default.
        """
        return self._get_bool("PEX_UNZIP")

    @defaulted_property(default=False)
    def PEX_VENV(self):
        # type: () -> bool
        """Boolean.

        Force this PEX to create a venv under $PEX_ROOT and re-execute from there.  If the pex file
        will be run multiple times under a stable $PEX_ROOT the venv creation will only be performed
        once and subsequent runs will enjoy lower startup latency.

        Default: false.
        """
        return self._get_bool("PEX_VENV")

    @defaulted_property(default=BinPath.FALSE)
    def PEX_VENV_BIN_PATH(self):
        # type: () -> BinPath.Value
        """String (false|prepend|append).

        When running in PEX_VENV mode, optionally add the scripts and console scripts of
        distributions in the PEX file to the $PATH.

        Default: false.
        """
        return BinPath.for_value(self._get_string("PEX_VENV_BIN_PATH"))

    @defaulted_property(default=False)
    def PEX_IGNORE_ERRORS(self):
        # type: () -> bool
        """Boolean.

        Ignore any errors resolving dependencies when invoking the PEX file. This can be useful if
        you know that a particular failing dependency is not necessary to run the application.

        Default: false.
        """
        return self._get_bool("PEX_IGNORE_ERRORS")

    @defaulted_property(default=InheritPath.FALSE)
    def PEX_INHERIT_PATH(self):
        # type: () -> InheritPath.Value
        """String (false|prefer|fallback)

        Allow inheriting packages from site-packages, user site-packages and the PYTHONPATH. By
        default, PEX scrubs any non stdlib packages from sys.path prior to invoking the application.
        Using 'prefer' causes PEX to shift any non-stdlib packages before the pex environment on
        sys.path and using 'fallback' shifts them after instead.

        Using this option is generally not advised, but can help in situations when certain
        dependencies do not conform to standard packaging practices and thus cannot be bundled into
        PEX files.

        See also PEX_EXTRA_SYS_PATH for how to *add* to the sys.path.

        Default: false.
        """
        try:
            return InheritPath.for_value(self._get_string("PEX_INHERIT_PATH"))
        except ValueError as e:
            die("Invalid value for PEX_INHERIT_PATH: {}".format(e))

    @defaulted_property(default=False)
    def PEX_INTERPRETER(self):
        # type: () -> bool
        """Boolean.

        Drop into a REPL instead of invoking the predefined entry point of this PEX. This can be
        useful for inspecting the PEX environment interactively.  It can also be used to treat the PEX
        file as an interpreter in order to execute other scripts in the context of the PEX file, e.g.
        "PEX_INTERPRETER=1 ./app.pex my_script.py".  Equivalent to setting PEX_MODULE to empty.

        Default: false.
        """
        return self._get_bool("PEX_INTERPRETER")

    @defaulted_property(default=False)
    def PEX_INTERPRETER_HISTORY(self):
        # type: () -> bool
        """Boolean.

        IF PEX_INTERPRETER is true, use a command history file for REPL user convenience.
        The location of the history file is determined by PEX_INTERPRETER_HISTORY_FILE.

        Note: Only supported on CPython interpreters.

        Default: false.
        """
        return self._get_bool("PEX_INTERPRETER_HISTORY")

    @defaulted_property(default=os.path.join("~", ".python_history"))
    def PEX_INTERPRETER_HISTORY_FILE(self):
        # type: () -> str
        """File.

        IF PEX_INTERPRETER_HISTORY is true, use this history file.
        The default is the standard CPython interpreter history location.

        Default: ~/.python_history.
        """
        return self._get_string("PEX_INTERPRETER_HISTORY_FILE")

    @property
    def PEX_MODULE(self):
        # type: () -> Optional[str]
        """String.

        Override the entry point into the PEX file.  Can either be a module, e.g.
        'SimpleHTTPServer', or a specific entry point in module:symbol form, e.g.  "myapp.bin:main".
        """
        return self._maybe_get_string("PEX_MODULE")

    @defaulted_property(default=False)
    def PEX_PROFILE(self):
        # type: () -> bool
        """Boolean.

        Enable application profiling.  If specified and PEX_PROFILE_FILENAME is not specified, PEX
        will print profiling information to stdout.
        """
        return self._get_bool("PEX_PROFILE")

    @property
    def PEX_PROFILE_FILENAME(self):
        # type: () -> Optional[str]
        """Filename.

        Profile the application and dump a profile into the specified filename in the standard
        "profile" module format.
        """
        return self._maybe_get_path("PEX_PROFILE_FILENAME")

    @defaulted_property(default="cumulative")
    def PEX_PROFILE_SORT(self):
        # type: () -> str
        """String.

        Toggle the profile sorting algorithm used to print out profile columns.

        Default: 'cumulative'.
        """
        return self._get_string("PEX_PROFILE_SORT")

    @property
    def PEX_PYTHON(self):
        # type: () -> Optional[str]
        """String.

        Override the Python interpreter used to invoke this PEX.  Can be either an absolute path to
        an interpreter or a base name e.g.  "python3.3".  If a base name is provided, the $PATH will
        be searched for an appropriate match.
        """
        return self._maybe_get_string("PEX_PYTHON")

    @property
    def PEX_PYTHON_PATH(self):
        # type: () -> Optional[Tuple[str, ...]]
        """String.

        A {pathsep!r} separated string containing paths of blessed Python interpreters for
        overriding the Python interpreter used to invoke this PEX. Can be absolute paths to
        interpreters or standard $PATH style directory entries that are searched for child files
        that are python binaries.

        For example, on a Unix system: "/path/to/python27:/path/to/python36-distribution/bin"
        """
        return self._maybe_get_path_tuple("PEX_PYTHON_PATH")

    @property
    def PEX_EXTRA_SYS_PATH(self):
        # type: () -> Tuple[str, ...]
        """String.

        A {pathsep!r} separated string containing paths to add to the runtime sys.path.

        Should be used sparingly, e.g., if you know that code inside this PEX needs to
        interact with code outside it.

        For example, on a Unix system: "/path/to/lib1:/path/to/lib2"

        This is distinct from PEX_INHERIT_PATH, which controls how the interpreter's
        existing sys.path (which you may not have control over) is scrubbed.

        See also PEX_PATH for how to merge packages from other pexes into the current environment.
        """.format(
            pathsep=os.pathsep
        )
        return self._maybe_get_path_tuple("PEX_EXTRA_SYS_PATH") or ()

    @defaulted_property(default=os.path.join("~", ".pex"))
    def PEX_ROOT(self):
        # type: () -> str
        """Directory.

        The directory location for PEX to cache any dependencies and code.  PEX must write not-zip-
        safe eggs and all wheels to disk in order to activate them.

        Default: ~/.pex
        """
        return self._get_path("PEX_ROOT")

    @PEX_ROOT.validator
    def _ensure_writeable_pex_root(self, raw_pex_root):
        pex_root = os.path.expanduser(raw_pex_root)
        if not can_write_dir(pex_root):
            tmp_root = os.path.realpath(safe_mkdtemp())
            pex_warnings.warn(
                "PEX_ROOT is configured as {pex_root} but that path is un-writeable, "
                "falling back to a temporary PEX_ROOT of {tmp_root} which will hurt "
                "performance.".format(pex_root=pex_root, tmp_root=tmp_root)
            )
            pex_root = self._environ["PEX_ROOT"] = tmp_root
        return pex_root

    @property
    def PEX_PATH(self):
        # type: () -> Tuple[str, ...]
        """A set of one or more PEX files.

        Merge the packages from other PEX files into the current environment.  This allows you to
        do things such as create a PEX file containing the "coverage" module or create PEX files
        containing plugin entry points to be consumed by a main application.  Paths should be
        specified in the same manner as $PATH. For example, on a Unix system
        PEX_PATH=/path/to/pex1.pex:/path/to/pex2.pex and so forth.

        See also PEX_EXTRA_SYS_PATH for how to add arbitrary entries to the sys.path.
        """
        return self._maybe_get_path_tuple("PEX_PATH", empty_string_is_cwd=False) or ()

    @property
    def PEX_SCRIPT(self):
        # type: () -> Optional[str]
        """String.

        The script name within the PEX environment to execute.  This must either be an entry point
        as defined in a distribution's console_scripts, or a script as defined in a distribution's
        scripts section.  While Python supports any script including shell scripts, PEX only
        supports invocation of Python scripts in this fashion.
        """
        return self._maybe_get_string("PEX_SCRIPT")

    @defaulted_property(default=False)
    def PEX_TEARDOWN_VERBOSE(self):
        # type: () -> bool
        """Boolean.

        Enable verbosity for when the interpreter shuts down.  This is mostly only useful for
        debugging PEX itself.

        Default: false.
        """
        return self._get_bool("PEX_TEARDOWN_VERBOSE")

    @defaulted_property(default=0)
    def PEX_VERBOSE(self):
        # type: () -> int
        """Integer.

        Set the verbosity level of PEX debug logging.  The higher the number, the more logging, with
        0 being disabled.  This environment variable can be extremely useful in debugging PEX
        environment issues.

        Default: 0
        """
        return self._get_int("PEX_VERBOSE")

    @defaulted_property(default=False)
    def PEX_IGNORE_RCFILES(self):
        # type: () -> bool
        """Boolean.

        Explicitly disable the reading/parsing of pexrc files (~/.pexrc).

        Default: false.
        """
        return self._get_bool("PEX_IGNORE_RCFILES")

    @property
    def PEX_EMIT_WARNINGS(self):
        # type: () -> Optional[bool]
        """Boolean.

        Emit UserWarnings to stderr. When false, warnings will only be logged at PEX_VERBOSE >= 1.
        When unset the build-time value of `--emit-warnings` will be used.

        Default: unset.
        """
        return self._maybe_get_bool("PEX_EMIT_WARNINGS")

    @defaulted_property(default=False)
    def PEX_TOOLS(self):
        # type: () -> bool
        """Boolean.

        Run the PEX tools.

        Default: false.
        """
        return self._get_bool("PEX_TOOLS")

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self._environ)


# Global singleton environment
ENV = Variables()


# TODO(John Sirois): Extract a runtime.modes package to hold code dealing with runtime mode
#  calculations: https://github.com/pantsbuild/pex/issues/1154
def _expand_pex_root(pex_root):
    # type: (str) -> str
    fallback = os.path.expanduser(pex_root)
    return os.path.expanduser(Variables.PEX_ROOT.value_or(ENV, fallback=fallback))


def unzip_dir(
    pex_root,  # type: str
    pex_hash,  # type: str
    expand_pex_root=True,  # type: bool
):
    # type: (...) -> str
    pex_root = _expand_pex_root(pex_root) if expand_pex_root else pex_root
    return os.path.join(pex_root, "unzipped_pexes", pex_hash)


def venv_dir(
    pex_file,  # type: str
    pex_root,  # type: str
    pex_hash,  # type: str
    has_interpreter_constraints,  # type: bool
    interpreter=None,  # type: Optional[PythonInterpreter]
    pex_path=(),  # type: Tuple[str, ...]
    expand_pex_root=True,  # type: bool
):
    # type: (...) -> str

    # The venv contents are affected by which PEX files are in play as well as which interpreter
    # is selected. The former is influenced via PEX_PATH and the latter is influenced by interpreter
    # constraints, PEX_PYTHON and PEX_PYTHON_PATH.

    pex_path_contents = {}  # type: Dict[str, Dict[str, str]]
    venv_contents = {"pex_path": pex_path_contents}  # type: Dict[str, Any]

    # PexInfo.pex_path and PEX_PATH are merged by PEX at runtime, so we include both and hash just
    # the distributions since those are the only items used from PEX_PATH adjoined PEX files; i.e.:
    # neither the entry_point nor any other PEX file data or metadata is used.
    def add_pex_path_items(pexes):
        # type: (Tuple[str, ...]) -> None
        if not pexes:
            return
        from pex.pex_info import PexInfo

        for pex in pexes:
            pex_path_contents[pex] = PexInfo.from_pex(pex).distributions

    add_pex_path_items(pex_path)
    add_pex_path_items(ENV.PEX_PATH)

    # There are two broad cases of interpreter selection, Pex chooses and user chooses. For the
    # former we hash Pex's selection criteria and for the latter we write down the actual
    # interpreter the user selected. N.B.: The pex_hash already includes an encoding of the
    # interpreter constraints, if any; so we don't hash those here again for the Pex chooses case.

    # This is relevant in all cases of interpreter selection and restricts the valid search path
    # for interpreter constraints, PEX_PYTHON and otherwise unconstrained PEXes.
    venv_contents["PEX_PYTHON_PATH"] = ENV.PEX_PYTHON_PATH

    interpreter_path = None  # type: Optional[str]
    precise_pex_python = ENV.PEX_PYTHON and os.path.exists(ENV.PEX_PYTHON)
    if precise_pex_python:
        # If there are interpreter_constraints and this interpreter doesn't meet them, then this
        # venv hash will never be used; so it's OK to hash based on the interpreter.
        interpreter_path = ENV.PEX_PYTHON
    elif ENV.PEX_PYTHON:
        # For an imprecise PEX_PYTHON that requires PATH lookup, PEX_PYTHON=python3.7, for example,
        # we just write down the constraint and let the PEX runtime determine the path of an
        # appropriate python3.7, if any (Pex chooses).
        venv_contents["PEX_PYTHON"] = ENV.PEX_PYTHON
    elif not has_interpreter_constraints:
        # Otherwise we can calculate the exact interpreter the PEX runtime will use. This ensures
        # ~unconstrained PEXes get a venv per interpreter used to invoke them with (user chooses).
        interpreter_binary = interpreter.binary if interpreter else sys.executable
        if (
            not ENV.PEX_PYTHON_PATH
            or interpreter_binary.startswith(ENV.PEX_PYTHON_PATH)
            or os.path.realpath(interpreter_binary).startswith(
                tuple(os.path.realpath(p) for p in ENV.PEX_PYTHON_PATH)
            )
        ):
            interpreter_path = interpreter_binary
    if interpreter_path:
        venv_contents["interpreter"] = os.path.realpath(interpreter_path)

    venv_contents_hash = hashlib.sha1(
        json.dumps(venv_contents, sort_keys=True).encode("utf-8")
    ).hexdigest()
    pex_root = _expand_pex_root(pex_root) if expand_pex_root else pex_root
    venv_path = os.path.join(pex_root, "venvs", pex_hash, venv_contents_hash)

    def warn(message):
        # type: (str) -> None
        from pex.pex_info import PexInfo

        pex_warnings.configure_warnings(ENV, PexInfo.from_pex(pex_file))
        pex_warnings.warn(message)

    if (
        ENV.PEX_PYTHON
        and not precise_pex_python
        and not re.match(r".*[^\d][\d]+\.[\d+]$", ENV.PEX_PYTHON)
    ):
        warn(
            dedent(
                """\
                Using a venv selected by PEX_PYTHON={pex_python} for {pex_file} at {venv_path}.

                If `{pex_python}` is upgraded or downgraded at some later date, this venv will still
                be used. To force re-creation of the venv using the upgraded or downgraded
                `{pex_python}` you will need to delete it at that point in time.

                To avoid this warning, either specify a Python binary with major and minor version
                in its name, like PEX_PYTHON=python{current_python_version} or else re-build the PEX
                with `--no-emit-warnings` or re-run the PEX with PEX_EMIT_WARNINGS=False.
                """.format(
                    pex_python=ENV.PEX_PYTHON,
                    pex_file=os.path.normpath(pex_file),
                    venv_path=venv_path,
                    current_python_version=".".join(map(str, sys.version_info[:2])),
                )
            )
        )
    if not interpreter_path and ENV.PEX_PYTHON_PATH:
        warn(
            dedent(
                """\
                Using a venv restricted by PEX_PYTHON_PATH={ppp} for {pex_file} at {venv_path}.

                If the contents of `{ppp}` changes at some later date, this venv and the interpreter
                selected from `{ppp}` will still be used. To force re-creation of the venv using
                the new pythons available on `{ppp}` you will need to delete it at that point in
                time.

                To avoid this warning, re-build the PEX with `--no-emit-warnings` or re-run the PEX
                with PEX_EMIT_WARNINGS=False.
                """
            ).format(
                ppp=os.pathsep.join(ENV.PEX_PYTHON_PATH),
                pex_file=os.path.normpath(pex_file),
                venv_path=venv_path,
            )
        )

    return venv_path



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import shutil
import subprocess
import sys
from textwrap import dedent

import pytest

from pex.common import temporary_dir
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import (
    InterpreterConstraints,
    UnsatisfiableInterpreterConstraintsError,
)
from pex.pex import PEX
from pex.pex_bootstrapper import (
    InterpreterTest,
    ensure_venv,
    find_compatible_interpreter,
    iter_compatible_interpreters,
)
from pex.pex_builder import PEXBuilder
from pex.testing import PY38, PY39, PY310, ensure_python_interpreter
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Any, Iterable, List, Optional


def basenames(*paths):
    # type: (*str) -> Iterable[str]
    return [os.path.basename(p) for p in paths]


def find_interpreters(
    path,  # type: Iterable[str]
    valid_basenames=None,  # type: Optional[Iterable[str]]
    constraints=(),  # type: Iterable[str]
    preferred_interpreter=None,  # type: Optional[PythonInterpreter]
):
    # type: (...) -> List[str]
    return [
        interp.binary
        for interp in iter_compatible_interpreters(
            path=tuple(path),
            valid_basenames=valid_basenames,
            interpreter_constraints=InterpreterConstraints.parse(*constraints),
            preferred_interpreter=preferred_interpreter,
        )
    ]


def test_find_compatible_interpreters():
    # type: () -> None
    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    py310 = ensure_python_interpreter(PY310)
    path = [py38, py39, py310]

    assert [py39, py310] == find_interpreters(path, constraints=[">3.9"])
    assert [py38] == find_interpreters(path, constraints=["<3.9"])

    assert [py310] == find_interpreters(path, constraints=[">{}".format(PY39)])
    assert [py39] == find_interpreters(path, constraints=[">{}, <{}".format(PY38, PY310)])
    assert [py310] == find_interpreters(path, constraints=[">=3.10"])

    with pytest.raises(UnsatisfiableInterpreterConstraintsError):
        find_interpreters(path, constraints=["<3"])

    with pytest.raises(UnsatisfiableInterpreterConstraintsError):
        find_interpreters(path, constraints=[">4"])

    with pytest.raises(UnsatisfiableInterpreterConstraintsError):
        find_interpreters(path, constraints=[">{}, <{}".format(PY38, PY39)])

    # All interpreters on PATH including whatever interpreter is currently running.
    all_known_interpreters = set(PythonInterpreter.from_binary(python) for python in path)
    all_known_interpreters.add(PythonInterpreter.get())

    interpreters = set(
        iter_compatible_interpreters(
            path, interpreter_constraints=InterpreterConstraints.parse("<3.9")
        )
    )
    i_rendered = "\n      ".join(sorted(map(repr, interpreters)))
    aki_rendered = "\n      ".join(sorted(map(repr, all_known_interpreters)))
    assert interpreters.issubset(all_known_interpreters), dedent(
        """
        interpreters '<3':
          {interpreters}
        
        all known interpreters:
          {all_known_interpreters}
        """.format(
            interpreters=i_rendered, all_known_interpreters=aki_rendered
        )
    )


def test_find_compatible_interpreters_none():
    # type: () -> None
    assert [] == find_interpreters([os.path.devnull])


def test_find_compatible_interpreters_none_with_valid_basenames():
    # type: () -> None
    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    path = [py38, py39]

    with pytest.raises(UnsatisfiableInterpreterConstraintsError) as exec_info:
        find_interpreters(path, valid_basenames=["python3.6"])

    exception_message = str(exec_info.value)
    assert py38 not in exception_message
    assert py39 not in exception_message


def test_find_compatible_interpreters_none_with_constraints():
    # type: () -> None
    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    path = [py38, py39]

    with pytest.raises(UnsatisfiableInterpreterConstraintsError) as exec_info:
        find_interpreters(path, constraints=[">=3.10"])

    exception_message = str(exec_info.value)
    assert py38 in exception_message
    assert py39 in exception_message
    assert ">=3.10" in exception_message


def test_find_compatible_interpreters_none_with_valid_basenames_and_constraints():
    # type: () -> None
    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    path = [py39, py38]

    with pytest.raises(UnsatisfiableInterpreterConstraintsError) as exec_info:
        find_interpreters(path, valid_basenames=basenames(py38), constraints=[">=3.9"])

    exception_message = str(exec_info.value)
    assert py38 in exception_message
    assert os.path.basename(py38) in exception_message
    assert py39 not in exception_message
    assert ">=3.9" in exception_message


def test_find_compatible_interpreters_with_valid_basenames():
    # type: () -> None
    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    py310 = ensure_python_interpreter(PY310)
    path = [py38, py39, py310]

    assert [py38] == find_interpreters(path, valid_basenames=basenames(py38))
    assert [py39, py310] == find_interpreters(
        path, valid_basenames=basenames(*reversed([py39, py310]))
    )


def test_find_compatible_interpreters_with_valid_basenames_and_constraints():
    # type: () -> None
    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    py310 = ensure_python_interpreter(PY310)
    path = [py38, py39, py310]

    assert [py39] == find_interpreters(
        path, valid_basenames=basenames(py38, py39), constraints=[">=3.9"]
    )


def test_find_compatible_interpreters_bias_current():
    # type: () -> None
    py310 = ensure_python_interpreter(PY310)
    current_interpreter = PythonInterpreter.get()
    assert [current_interpreter.binary, py310] == find_interpreters([py310, sys.executable])
    assert [current_interpreter.binary, py310] == find_interpreters([sys.executable, py310])


def test_find_compatible_interpreters_siblings_of_current_issues_1109():
    # type: () -> None
    py39 = ensure_python_interpreter(PY39)
    py310 = ensure_python_interpreter(PY310)

    with temporary_dir() as path_entry:
        python39 = os.path.join(path_entry, "python3.9")
        shutil.copy(py39, python39)

        python310 = os.path.join(path_entry, "python3.10")
        shutil.copy(py310, python310)

        assert [os.path.realpath(p) for p in (python310, python39)] == find_interpreters(
            path=[path_entry], preferred_interpreter=PythonInterpreter.from_binary(python310)
        )


def test_ensure_venv_activate_issues_1276(tmpdir):
    # type: (Any) -> None
    pex_root = os.path.join(str(tmpdir), "pex_root")
    pb = PEXBuilder()
    pb.info.pex_root = pex_root
    pb.info.venv = True
    pb.freeze()

    venv_pex = ensure_venv(PEX(pb.path()))

    expected_python_bin_dir = (
        subprocess.check_output(
            args=[
                venv_pex.pex,
                "-c",
                "import os, sys; print(os.path.realpath(os.path.dirname(sys.executable)))",
            ],
        )
        .decode("utf-8")
        .strip()
    )
    bash_activate = venv_pex.bin_file("activate")

    subprocess.check_call(
        args=[
            "/usr/bin/env",
            "bash",
            "-c",
            dedent(
                """\
                source "{activate_script}"

                actual_python_bin_dir="$(dirname "$(command -v python)")"
                expected_python_bin_dir="{expected_python_bin_dir}"

                if [ "${{actual_python_bin_dir}}" != "${{expected_python_bin_dir}}" ]; then
                    echo >&2 "Actual Python Bin Dir: ${{actual_python_bin_dir}}"
                    echo >&2 "Expected Python Bin Dir: ${{expected_python_bin_dir}}"
                  exit 42
                fi
                """.format(
                    activate_script=bash_activate, expected_python_bin_dir=expected_python_bin_dir
                )
            ),
        ]
    )


def test_pp_invalid():
    # type: () -> None
    with ENV.patch(PEX_PYTHON="/invalid/abs/python"):
        with pytest.raises(
            UnsatisfiableInterpreterConstraintsError,
            match=(
                r"The specified PEX_PYTHON=/invalid/abs/python could not be identified as a "
                r"valid Python interpreter."
            ),
        ):
            find_compatible_interpreter()


def test_pp_exact():
    # type: () -> None
    py310 = ensure_python_interpreter(PY310)
    with ENV.patch(PEX_PYTHON=py310):
        assert PythonInterpreter.from_binary(py310) == find_compatible_interpreter()


def test_pp_exact_on_ppp():
    # type: () -> None

    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    py310 = ensure_python_interpreter(PY310)

    with ENV.patch(
        PEX_PYTHON=py310,
        PEX_PYTHON_PATH=":".join(os.path.dirname(py) for py in (py38, py39, py310)),
    ):
        assert PythonInterpreter.from_binary(py310) == find_compatible_interpreter()


def test_pp_exact_satisfies_constraints():
    # type: () -> None

    py310 = ensure_python_interpreter(PY310)

    pb = PEXBuilder()
    pb.info.interpreter_constraints = InterpreterConstraints.parse(">=3.7")
    pb.freeze()

    with ENV.patch(PEX_PYTHON=py310):
        assert PythonInterpreter.from_binary(py310) == find_compatible_interpreter(
            interpreter_test=InterpreterTest(entry_point=pb.path(), pex_info=pb.info),
        )


def test_pp_exact_does_not_satisfy_constraints():
    # type: () -> None

    py310 = ensure_python_interpreter(PY310)

    pb = PEXBuilder()
    pb.info.interpreter_constraints = InterpreterConstraints.parse("<=3.7")
    pb.freeze()

    with ENV.patch(PEX_PYTHON=py310):
        with pytest.raises(
            UnsatisfiableInterpreterConstraintsError,
            match=r"Failed to find a compatible PEX_PYTHON={pp}.".format(pp=py310),
        ):
            find_compatible_interpreter(
                interpreter_test=InterpreterTest(entry_point=pb.path(), pex_info=pb.info),
            )


def test_pp_exact_not_on_ppp():
    # type: () -> None

    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    py310 = ensure_python_interpreter(PY310)

    with ENV.patch(
        PEX_PYTHON=py310, PEX_PYTHON_PATH=":".join(os.path.dirname(py) for py in (py38, py39))
    ):
        with pytest.raises(
            UnsatisfiableInterpreterConstraintsError,
            match=r"The specified PEX_PYTHON={pp} did not meet other constraints.".format(pp=py310),
        ):
            find_compatible_interpreter()



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import platform
import subprocess
from contextlib import contextmanager
from textwrap import dedent

import pytest

from pex import resolver
from pex.common import temporary_dir
from pex.compatibility import to_bytes
from pex.dist_metadata import DistMetadata, Distribution
from pex.environment import PEXEnvironment, _InvalidWheelName, _RankedDistribution
from pex.fingerprinted_distribution import FingerprintedDistribution
from pex.inherit_path import InheritPath
from pex.interpreter import PythonInterpreter
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pex import PEX
from pex.pex_builder import PEXBuilder
from pex.pex_info import PexInfo
from pex.targets import LocalInterpreter, Targets
from pex.testing import (
    IS_LINUX_X86_64,
    IS_PYPY3,
    PY38,
    WheelBuilder,
    ensure_python_interpreter,
    install_wheel,
    make_bdist,
    temporary_content,
    temporary_filename,
)
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Tuple


@contextmanager
def yield_pex_builder(zip_safe=True, interpreter=None):
    # type: (bool, Optional[PythonInterpreter]) -> Iterator[PEXBuilder]
    with temporary_dir() as td, make_bdist("p1", zip_safe=zip_safe, interpreter=interpreter) as p1:
        pb = PEXBuilder(path=td, interpreter=interpreter)
        pb.add_dist_location(p1.location)
        yield pb


def assert_force_local_implicit_ns_packages_issues_598(
    interpreter=None, requirements=(), create_ns_packages=True
):
    def create_foo_bar_setup(name, **extra_args):
        # type: (str, **Any) -> str
        setup_args = dict(name=name, version="0.0.1", packages=["foo", "foo.bar"])
        if create_ns_packages:
            setup_args.update(namespace_packages=["foo", "foo.bar"])
        if requirements:
            setup_args.update(install_requires=list(requirements))
        setup_args.update(extra_args)

        return dedent(
            """
            from setuptools import setup
            
            setup(**{setup_args!r})
            """.format(
                setup_args=setup_args
            )
        )

    def with_foo_bar_ns_packages(content):
        # type: (Dict[str, str]) -> Dict[str, str]
        ns_packages = (
            {
                os.path.join(
                    pkg, "__init__.py"
                ): '__import__("pkg_resources").declare_namespace(__name__)'
                for pkg in ("foo", "foo/bar")
            }
            if create_ns_packages
            else {}
        )
        ns_packages.update(content)
        return ns_packages

    content1 = with_foo_bar_ns_packages(
        {
            "foo/bar/spam.py": "identify = lambda: 42",
            "setup.py": create_foo_bar_setup("foo-bar-spam"),
        }
    )

    content2 = with_foo_bar_ns_packages(
        {
            "foo/bar/eggs.py": dedent(
                """
                # NB: This only works when this content is unpacked loose on the filesystem!
                def read_self():
                    with open(__file__) as fp:
                        return fp.read()
                """
            )
        }
    )

    content3 = with_foo_bar_ns_packages(
        {
            "foobaz": dedent(
                """\
                #!python
                import sys
                
                from foo.bar import baz
                
                sys.exit(baz.main())
                """
            ),
            "foo/bar/baz.py": dedent(
                """
                import sys
                
                from foo.bar import eggs, spam
                
                def main():
                    assert len(eggs.read_self()) > 0
                    return spam.identify()
                """
            ),
            "setup.py": create_foo_bar_setup("foo-bar-baz", scripts=["foobaz"]),
        }
    )

    def add_requirements(builder, cache):
        # type: (PEXBuilder, str) -> None
        for installed_dist in resolver.resolve(
            targets=Targets(interpreters=(builder.interpreter,)),
            requirements=requirements,
        ).installed_distributions:
            builder.add_distribution(installed_dist.distribution)
            for direct_req in installed_dist.direct_requirements:
                builder.add_requirement(direct_req)

    def add_wheel(builder, content):
        # type: (PEXBuilder, Dict[str, str]) -> None
        with temporary_content(content) as project:
            dist = install_wheel(WheelBuilder(project, interpreter=builder.interpreter).bdist())
            builder.add_dist_location(dist.location)

    def add_sources(builder, content):
        # type: (PEXBuilder, Dict[str, str]) -> None
        with temporary_content(content) as project:
            for path in content.keys():
                builder.add_source(os.path.join(project, path), path)

    with temporary_dir() as root, temporary_dir() as cache:
        pex_info1 = PexInfo.default()
        pex1 = os.path.join(root, "pex1.pex")
        builder1 = PEXBuilder(interpreter=interpreter, pex_info=pex_info1)
        add_requirements(builder1, cache)
        add_wheel(builder1, content1)
        add_sources(builder1, content2)
        builder1.build(pex1)

        pex_info2 = PexInfo.default()
        pex_info2.pex_path = [pex1]
        pex2 = os.path.join(root, "pex2")
        builder2 = PEXBuilder(path=pex2, interpreter=interpreter, pex_info=pex_info2)
        add_requirements(builder2, cache)
        add_wheel(builder2, content3)
        builder2.set_script("foobaz")
        builder2.freeze()

        assert 42 == PEX(pex2, interpreter=interpreter).run()


def get_setuptools_requirement(interpreter=None):
    # type: (Optional[PythonInterpreter]) -> str
    # We use a very old version of setuptools to prove the point the user version is what is used
    # here and not the vendored version (when possible). A newer setuptools is needed though to work
    # with python 3.
    return (
        "setuptools==1.0"
        if (interpreter or PythonInterpreter.get()).version[0] == 2
        else "setuptools==17.0"
    )


@pytest.mark.xfail(IS_PYPY3, reason="https://github.com/pantsbuild/pex/issues/1210")
def test_issues_598_explicit_any_interpreter():
    # type: () -> None
    assert_force_local_implicit_ns_packages_issues_598(
        requirements=[get_setuptools_requirement()], create_ns_packages=True
    )


def test_issues_598_explicit_missing_requirement():
    # type: () -> None
    assert_force_local_implicit_ns_packages_issues_598(create_ns_packages=True)


@pytest.fixture
def python_38_interpreter():
    # type: () -> PythonInterpreter
    # Python 3.7 supports implicit namespace packages.
    return PythonInterpreter.from_binary(ensure_python_interpreter(PY38))


def test_issues_598_implicit(python_38_interpreter):
    # type: (PythonInterpreter) -> None
    assert_force_local_implicit_ns_packages_issues_598(
        interpreter=python_38_interpreter, create_ns_packages=False
    )


def test_issues_598_implicit_explicit_mixed(python_38_interpreter):
    # type: (PythonInterpreter) -> None
    assert_force_local_implicit_ns_packages_issues_598(
        interpreter=python_38_interpreter,
        requirements=[get_setuptools_requirement(python_38_interpreter)],
        create_ns_packages=True,
    )


_KNOWN_BAD_APPLE_INTERPRETER = (
    "/System/Library/Frameworks/Python.framework/Versions/"
    "2.7/Resources/Python.app/Contents/MacOS/Python"
)


@pytest.mark.skipif(
    not os.path.exists(_KNOWN_BAD_APPLE_INTERPRETER)
    or subprocess.check_output(
        [
            _KNOWN_BAD_APPLE_INTERPRETER,
            "-c" "import sys; print('.'.join(map(str, sys.version_info[:3])))",
        ]
    )
    != b"2.7.10",
    reason="Test requires known bad Apple interpreter {}".format(_KNOWN_BAD_APPLE_INTERPRETER),
)
def test_osx_platform_intel_issue_523():
    # type: () -> None

    def bad_interpreter():
        # type: () -> PythonInterpreter
        return PythonInterpreter.from_binary(_KNOWN_BAD_APPLE_INTERPRETER)

    # We need to run the bad interpreter with a modern, non-Apple-Extras setuptools in order to
    # successfully install psutil; yield_pex_builder sets up the bad interpreter with our vendored
    # setuptools and wheel extras.
    with yield_pex_builder(interpreter=bad_interpreter()) as pb, temporary_filename() as pex_file:
        for installed_dist in resolver.resolve(
            targets=Targets(interpreters=(pb.interpreter,)),
            requirements=["psutil==5.4.3"],
        ).installed_distributions:
            pb.add_dist_location(installed_dist.distribution.location)
        pb.build(pex_file)

        # NB: We want PEX to find the bare bad interpreter at runtime.
        pex = PEX(pex_file, interpreter=bad_interpreter())

        def run(args, **env):
            # type: (Iterable[str], **str) -> Tuple[int, str, str]
            pex_env = os.environ.copy()
            pex_env["PEX_VERBOSE"] = "1"
            pex_env.update(**env)
            process = pex.run(
                args=args,
                env=pex_env,
                blocking=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = process.communicate()
            return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")

        returncode, _, stderr = run(["-c", "import psutil"])
        assert 0 == returncode, "Process failed with exit code {} and stderr:\n{}".format(
            returncode, stderr
        )

        returncode, stdout, stderr = run(["-c", "import pkg_resources"])
        assert 0 != returncode, (
            "Isolated pex process succeeded but should not have found pkg-resources:\n"
            "STDOUT:\n"
            "{}\n"
            "STDERR:\n"
            "{}".format(stdout, stderr)
        )

        returncode, stdout, stderr = run(
            ["-c", "import pkg_resources; print(pkg_resources.get_supported_platform())"],
            # Let the bad interpreter site-packages setuptools leak in.
            PEX_INHERIT_PATH=InheritPath.for_value(True).value,
        )
        assert 0 == returncode, "Process failed with exit code {} and stderr:\n{}".format(
            returncode, stderr
        )

        # Verify this worked along side the previously problematic pkg_resources-reported platform.
        release, _, _ = platform.mac_ver()
        major_minor = ".".join(release.split(".")[:2])
        machine = platform.machine()
        assert "macosx-{}-{}".format(major_minor, machine) == stdout.strip()


def test_activate_extras_issue_615():
    # type: () -> None
    with yield_pex_builder() as pb:
        for installed_dist in resolver.resolve(
            targets=Targets(interpreters=(pb.interpreter,)),
            requirements=["pex[requests]==1.6.3"],
        ).installed_distributions:
            for direct_req in installed_dist.direct_requirements:
                pb.add_requirement(direct_req)
            pb.add_dist_location(installed_dist.distribution.location)
        pb.set_script("pex")
        pb.freeze()
        process = PEX(pb.path(), interpreter=pb.interpreter).run(
            args=["--version"],
            env={"PEX_VERBOSE": "9"},
            blocking=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()
        assert 0 == process.returncode, "Process failed with exit code {} and output:\n{}".format(
            process.returncode, stderr
        )
        assert to_bytes("{} 1.6.3".format(os.path.basename(pb.path()))) == stdout.strip()


def assert_namespace_packages_warning(distribution, version, expected_warning):
    # type: (str, str, bool) -> None
    requirement = "{}=={}".format(distribution, version)
    pb = PEXBuilder()
    for installed_dist in resolver.resolve(requirements=[requirement]).installed_distributions:
        pb.add_dist_location(installed_dist.distribution.location)
    pb.freeze()

    process = PEX(pb.path()).run(args=["-c", ""], blocking=False, stderr=subprocess.PIPE)
    _, stderr = process.communicate()
    stderr_text = stderr.decode("utf8")

    partial_warning_preamble = "PEXWarning: The `pkg_resources` package was loaded"
    partial_warning_detail = "{} namespace packages:".format(requirement)

    if expected_warning:
        assert partial_warning_preamble in stderr_text
        assert partial_warning_detail in stderr_text
    else:
        assert partial_warning_preamble not in stderr_text
        assert partial_warning_detail not in stderr_text


def test_present_non_empty_namespace_packages_metadata_does_warn():
    # type: () -> None
    assert_namespace_packages_warning("twitter.common.lang", "0.3.11", expected_warning=True)


def test_present_but_empty_namespace_packages_metadata_does_not_warn():
    # type: () -> None
    assert_namespace_packages_warning("pycodestyle", "2.5.0", expected_warning=False)


def create_dist(
    location,  # type: str
    project_name="foo",  # type: str
    version="1.0.0",  # type: str
):
    # type: (...) -> FingerprintedDistribution
    return FingerprintedDistribution(
        distribution=Distribution(
            location=location,
            metadata=DistMetadata(project_name=ProjectName(project_name), version=Version(version)),
        ),
        fingerprint=location,
    )


@pytest.fixture
def cpython_38_environment(python_38_interpreter):
    return PEXEnvironment(
        pex="",
        pex_info=PexInfo.default(),
        target=LocalInterpreter.create(python_38_interpreter),
    )


@pytest.mark.parametrize(
    ("wheel_distribution", "wheel_is_linux_x86_64"),
    [
        pytest.param(
            create_dist("llvmlite-0.29.0-cp38-cp38-linux_x86_64.whl", "llvmlite", "0.29.0"),
            True,
            id="without_build_tag_linux",
        ),
        pytest.param(
            create_dist("llvmlite-0.29.0-1-cp38-cp38-linux_x86_64.whl", "llvmlite", "0.29.0"),
            True,
            id="with_build_tag_linux",
        ),
        pytest.param(
            create_dist("llvmlite-0.29.0-cp38-cp38-macosx_10.9_x86_64.whl", "llvmlite", "0.29.0"),
            False,
            id="without_build_tag_osx",
        ),
        pytest.param(
            create_dist("llvmlite-0.29.0-1-cp38-cp38-macosx_10.9_x86_64.whl", "llvmlite", "0.29.0"),
            False,
            id="with_build_tag_osx",
        ),
    ],
)
def test_can_add_handles_optional_build_tag_in_wheel(
    cpython_38_environment, wheel_distribution, wheel_is_linux_x86_64
):
    # type: (PEXEnvironment, FingerprintedDistribution, bool) -> None
    native_wheel = IS_LINUX_X86_64 and wheel_is_linux_x86_64
    added = isinstance(cpython_38_environment._can_add(wheel_distribution), _RankedDistribution)
    assert added is native_wheel


def test_can_add_handles_invalid_wheel_filename(cpython_38_environment):
    # type: (PEXEnvironment) -> None
    dist = create_dist("pep427-invalid.whl")
    assert _InvalidWheelName(dist, "pep427-invalid") == cpython_38_environment._can_add(dist)


@pytest.fixture
def assert_cpython_38_environment_can_add(cpython_38_environment):
    # type: (PEXEnvironment) -> Callable[[FingerprintedDistribution], _RankedDistribution]
    def assert_can_add(fingerprinted_dist):
        # type: (FingerprintedDistribution) -> _RankedDistribution
        rank = cpython_38_environment._can_add(fingerprinted_dist)
        assert isinstance(rank, _RankedDistribution)
        return rank

    return assert_can_add


# So the following tests pass on all platforms.
_wheel_tags = "macosx_10_9_x86_64.macosx_11_0_arm64.linux_x86_64.linux_aarch64"


def test_can_add_ranking_platform_tag_more_specific(assert_cpython_38_environment_can_add):
    # type: (Callable[[FingerprintedDistribution], _RankedDistribution]) -> None
    ranked_specific = assert_cpython_38_environment_can_add(
        create_dist("foo-1.0.0-cp38-cp38-{}.whl".format(_wheel_tags), "foo", "1.0.0")
    )
    ranked_universal = assert_cpython_38_environment_can_add(
        create_dist("foo-2.0.0-py2.py3-none-any.whl", "foo", "2.0.0")
    )
    assert ranked_specific < ranked_universal

    ranked_almost_py3universal = assert_cpython_38_environment_can_add(
        create_dist("foo-2.0.0-py3-none-any.whl", "foo", "2.0.0")
    )
    assert ranked_universal.rank == ranked_almost_py3universal.rank, (
        "Expected the 'universal' compressed tag set to be expanded into two tags and the more "
        "specific tag picked from those two for ranking."
    )


def test_can_add_ranking_version_newer_tie_break(assert_cpython_38_environment_can_add):
    # type: (Callable[[FingerprintedDistribution], _RankedDistribution]) -> None
    ranked_v1 = assert_cpython_38_environment_can_add(
        create_dist("foo-1.0.0-cp38-cp38-{}.whl".format(_wheel_tags), "foo", "1.0.0")
    )
    ranked_v2 = assert_cpython_38_environment_can_add(
        create_dist("foo-2.0.0-cp38-cp38-{}.whl".format(_wheel_tags), "foo", "2.0.0")
    )
    assert ranked_v2 < ranked_v1



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import logging
from argparse import ArgumentParser

from pex import pex_bootstrapper
from pex.commands.command import JsonMixin, OutputMixin
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import (
    InterpreterConstraint,
    InterpreterConstraints,
    UnsatisfiableInterpreterConstraintsError,
)
from pex.pex import PEX
from pex.pex_bootstrapper import InterpreterTest
from pex.result import Error, Ok, Result
from pex.tools.command import PEXCommand
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Any, Dict, Iterator

logger = logging.getLogger(__name__)


class Interpreter(JsonMixin, OutputMixin, PEXCommand):
    """Prints the path of the preferred interpreter to run the given PEX file with, if any."""

    @classmethod
    def add_arguments(cls, parser):
        # type: (ArgumentParser) -> None
        cls.add_output_option(parser, entity="Python interpreter path"),
        parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="Print all compatible interpreters, preferred first.",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help=(
                "Provide more information about the interpreter in json format. "
                "Once: include the interpreter requirement and platform in addition to its path. "
                "Twice: include the interpreter's supported tags. "
                "Thrice: include the interpreter's environment markers and its venv affiliation, "
                "if any."
            ),
        )
        cls.add_json_options(parser, entity="verbose output")
        cls.register_global_arguments(parser, include_verbosity=False)

    def _find_interpreters(self, pex):
        # type: (PEX) -> Iterator[PythonInterpreter]
        if not self.options.all:
            yield pex.interpreter
            return

        if ENV.PEX_PYTHON:
            logger.warning(
                "Ignoring PEX_PYTHON={} in order to scan for all compatible "
                "interpreters.".format(ENV.PEX_PYTHON)
            )
        pex_info = pex.pex_info()
        for interpreter in pex_bootstrapper.iter_compatible_interpreters(
            path=ENV.PEX_PYTHON_PATH,
            interpreter_constraints=pex_info.interpreter_constraints,
            interpreter_test=InterpreterTest(entry_point=pex.path(), pex_info=pex_info),
        ):
            yield interpreter

    def run(self, pex):
        # type: (PEX) -> Result
        if self.options.indent and not self.options.verbose:
            logger.warning(
                "Ignoring --indent={} since --verbose mode is not enabled.".format(
                    self.options.indent
                )
            )
        with self.output(self.options) as out:
            try:
                for interpreter in self._find_interpreters(pex):
                    if self.options.verbose:
                        interpreter_info = {
                            "path": interpreter.binary,
                            "requirement": str(InterpreterConstraint.exact_version(interpreter)),
                            "platform": str(interpreter.platform),
                        }  # type: Dict[str, Any]
                        if self.options.verbose >= 2:
                            interpreter_info[
                                "supported_tags"
                            ] = interpreter.identity.supported_tags.to_string_list()
                        if self.options.verbose >= 3:
                            interpreter_info[
                                "env_markers"
                            ] = interpreter.identity.env_markers.as_dict()
                            interpreter_info["venv"] = interpreter.is_venv
                            if interpreter.is_venv:
                                interpreter_info[
                                    "base_interpreter"
                                ] = interpreter.resolve_base_interpreter().binary
                        self.dump_json(self.options, interpreter_info, out)
                    else:
                        out.write(interpreter.binary)
                    out.write("\n")
            except UnsatisfiableInterpreterConstraintsError as e:
                return Error(str(e))

        return Ok()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import shutil
import subprocess

import pytest

from pex.dist_metadata import Distribution
from pex.interpreter import PythonInterpreter
from pex.pep_503 import ProjectName
from pex.testing import ALL_PY_VERSIONS, VenvFactory, all_python_venvs, ensure_python_venv
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import InvalidVirtualenvError, Virtualenv

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Tuple


def test_invalid(tmpdir):
    # type: (Any) -> None

    with pytest.raises(InvalidVirtualenvError):
        Virtualenv(venv_dir=str(tmpdir))

    venv_dir = os.path.join(str(tmpdir), "venv")
    Virtualenv.create(venv_dir=venv_dir)
    venv = Virtualenv(venv_dir=venv_dir)

    shutil.rmtree(venv.site_packages_dir)
    with pytest.raises(InvalidVirtualenvError):
        Virtualenv(venv_dir=venv_dir)


def test_enclosing(tmpdir):
    # type: (Any) -> None

    venv_dir = os.path.join(str(tmpdir), "venv")
    venv = Virtualenv.create(venv_dir=venv_dir)

    enclosing = Virtualenv.enclosing(venv.interpreter)
    assert enclosing is not None
    assert venv_dir == enclosing.venv_dir

    enclosing = Virtualenv.enclosing(venv.interpreter.binary)
    assert enclosing is not None
    assert venv_dir == enclosing.venv_dir

    assert Virtualenv.enclosing(venv.interpreter.resolve_base_interpreter()) is None


def index_distributions(venv):
    # type: (Virtualenv) -> Dict[ProjectName, Distribution]
    return {dist.metadata.project_name: dist for dist in venv.iter_distributions()}


def test_iter_distributions_setuptools_not_leaked(tmpdir):
    # type: (Any) -> None

    empty_venv_dir = os.path.join(str(tmpdir), "empty.venv")
    empty_venv = Virtualenv.create(venv_dir=empty_venv_dir)
    dists = index_distributions(empty_venv)
    assert ProjectName("setuptools") not in dists


@pytest.mark.parametrize(
    "venv_factory",
    [
        pytest.param(venv_factory, id=venv_factory.python_version)
        for venv_factory in all_python_venvs()
    ],
)
def test_iter_distributions(venv_factory):
    # type: (VenvFactory) -> None

    python, pip = venv_factory.create_venv()
    venv = Virtualenv.enclosing(python)
    assert venv is not None

    dists = index_distributions(venv)
    pip_dist = dists.get(ProjectName("pip"))
    assert pip_dist is not None, "Expected venv to have Pip installed."
    assert os.path.realpath(venv.site_packages_dir) == os.path.realpath(pip_dist.location)
    assert ProjectName("cowsay") not in dists

    subprocess.check_call(args=[pip, "install", "cowsay==4.0"])
    dists = index_distributions(venv)
    cowsay_dist = dists.get(ProjectName("cowsay"))
    assert cowsay_dist is not None, "Expected venv to have cowsay installed."
    assert "4.0" == cowsay_dist.version
    assert os.path.realpath(venv.site_packages_dir) == os.path.realpath(cowsay_dist.location)


def test_iter_distributions_spaces(tmpdir):
    # type: (Any) -> None

    venv_dir = os.path.join(str(tmpdir), "face palm")
    venv = Virtualenv.create(venv_dir=venv_dir)
    dists = index_distributions(venv)
    pip_dist = dists.get(ProjectName("pip"))
    assert pip_dist is None, "Expected venv to not have Pip installed."

    venv.install_pip()
    dists = index_distributions(venv)
    pip_dist = dists.get(ProjectName("pip"))
    assert pip_dist is not None, "Expected venv to have Pip installed."
    assert os.path.realpath(venv.site_packages_dir) == os.path.realpath(pip_dist.location)



# coding=utf-8
# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class NetworkConfiguration(object):
    """Configuration for network requests made by the resolver.

    :param retries: The maximum number of retries each connection should attempt.
    :param timeout: The socket timeout in seconds.
    :param proxy: A proxy to use in the form [user:passwd@]proxy.server:port.
    :param cert: The path to an alternate CA bundle.
    :param client_cert: The path to an SSL client certificate which should be a single file
                        containing the private key and the certificate in PEM format.
    :return: The resulting network configuration.
    """

    retries = attr.ib(default=5)  # type: int
    timeout = attr.ib(default=15)  # type: float
    proxy = attr.ib(default=None)  # type: Optional[str]
    cert = attr.ib(default=None)  # type: Optional[str]
    client_cert = attr.ib(default=None)  # type: Optional[str]

    @retries.validator
    def _validate_retries(self, attribute, value):
        if value < 0:
            raise ValueError(
                "The {} parameter should be >= 0; given: {}".format(attribute.name, value)
            )

    @timeout.validator
    def _validate_timeout(self, attribute, value):
        if value <= 0:
            raise ValueError(
                "The {} parameter should be > 0; given: {}".format(attribute.name, value)
            )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import glob
import os
import subprocess

import pytest

from pex.common import temporary_dir
from pex.pip.installation import get_pip
from pex.testing import IS_LINUX_ARM64, IS_PYPY, PY_VER, run_pex_command


@pytest.mark.skipif(
    IS_LINUX_ARM64 or IS_PYPY or PY_VER > (3, 10),
    reason=(
        "The cryptography 2.6.1 project only has pre-built wheels for CPython <= 3.10 for some "
        "platforms available on PyPI and this test relies upon a pre-built wheel being available."
    ),
)
def test_abi3_resolution():
    # type: () -> None
    # The cryptography team releases the following relevant pre-built wheels for version 2.6.1:
    # cryptography-2.6.1-cp27-cp27m-macosx_10_6_intel.whl
    # cryptography-2.6.1-cp27-cp27m-manylinux1_x86_64.whl
    # cryptography-2.6.1-cp27-cp27mu-manylinux1_x86_64.whl
    # cryptography-2.6.1-cp34-abi3-macosx_10_6_intel.whl
    # cryptography-2.6.1-cp34-abi3-manylinux1_x86_64.whl
    # With pex in --no-build mode, we force a test that pex abi3 resolution works when this test is
    # run under CPython>3.4,<3.11 on OSX and linux.

    with temporary_dir() as td:
        # The dependency graph for cryptography-2.6.1 includes pycparser which is only released as an
        # sdist. Since we want to test in --no-build, we pre-resolve/build the pycparser wheel here and
        # add the resulting wheelhouse to the --no-build pex command.
        download_dir = os.path.join(td, ".downloads")
        get_pip().spawn_download_distributions(
            download_dir=download_dir, requirements=["pycparser"]
        ).wait()
        wheel_dir = os.path.join(td, ".wheels")
        get_pip().spawn_build_wheels(
            wheel_dir=wheel_dir, distributions=glob.glob(os.path.join(download_dir, "*"))
        ).wait()

        cryptography_pex = os.path.join(td, "cryptography.pex")
        res = run_pex_command(
            ["-f", wheel_dir, "--no-build", "cryptography==2.6.1", "-o", cryptography_pex]
        )
        res.assert_success()

        subprocess.check_call([cryptography_pex, "-c", "import cryptography"])



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from argparse import ArgumentParser

from pex.commands.command import JsonMixin, OutputMixin
from pex.pex import PEX
from pex.result import Ok, Result
from pex.tools.command import PEXCommand


class Info(JsonMixin, OutputMixin, PEXCommand):
    """Dumps the PEX-INFO json contained in a PEX file."""

    @classmethod
    def add_arguments(cls, parser):
        # type: (ArgumentParser) -> None
        cls.add_output_option(parser, entity="PEX-INFO json")
        cls.add_json_options(parser, entity="PEX-INFO")
        cls.register_global_arguments(parser)

    def run(self, pex):
        # type: (PEX) -> Result
        with self.output(self.options) as out:
            self.dump_json(self.options, pex.pex_info().as_json_dict(), out)
            out.write("\n")
        return Ok()



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import itertools
import json
import os


def patch():
    from pip._internal.utils import compatibility_tags  # type: ignore[import]
    from pip._vendor.packaging import tags  # type: ignore[import]

    # N.B.: The following environment variable is used by the Pex runtime to control Pip and must be
    # kept in-sync with `__init__.py`.
    patched_tags_file = os.environ.pop("_PEX_PATCHED_TAGS_FILE")
    with open(patched_tags_file) as fp:
        patched_tags = tuple(
            itertools.chain.from_iterable(tags.parse_tag(tag) for tag in json.load(fp))
        )

    def get_supported(*_args, **_kwargs):
        return list(patched_tags)

    compatibility_tags.get_supported = get_supported



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import sys

import pytest

from pex.cli.testing import run_pex3
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    sys.version_info[:2] < (3, 7),
    reason="The tested distribution is only compatible with Python >= 3.7",
)
def test_get_requires_for_build_wheel(tmpdir):
    # type: (Any) -> None

    run_pex3("lock", "create", "cairocffi==1.5.1").assert_success()



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
import os

import pytest

from pex.common import safe_mkdtemp
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import InterpreterConstraint
from pex.pex_builder import PEXBuilder
from pex.testing import PY38, PY310, ensure_python_interpreter
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@pytest.fixture(scope="module")
def python38():
    # type: () -> PythonInterpreter
    return PythonInterpreter.from_binary(ensure_python_interpreter(PY38))


@pytest.fixture(scope="module")
def python310():
    # type: () -> PythonInterpreter
    return PythonInterpreter.from_binary(ensure_python_interpreter(PY310))


@attr.s(frozen=True)
class InterpreterTool(object):
    tools_pex = attr.ib()  # type: str
    interpreter = attr.ib()  # type: PythonInterpreter
    other_interpreters = attr.ib(default=())  # type: Iterable[PythonInterpreter]

    @classmethod
    def create(
        cls,
        interpreter,  # type: PythonInterpreter
        *other_interpreters  # type: PythonInterpreter
    ):
        # type: (...) -> InterpreterTool
        pex_builder = PEXBuilder(interpreter=interpreter)
        pex_builder.info.includes_tools = True
        pex_builder.freeze()
        return cls(
            tools_pex=pex_builder.path(),
            interpreter=interpreter,
            other_interpreters=other_interpreters,
        )

    def run(
        self,
        *args,  # type: str
        **env  # type: str
    ):
        # type: (...) -> str
        cmd = [self.tools_pex, "interpreter"]
        if args:
            cmd.extend(args)

        environ = os.environ.copy()
        interpreters = [self.interpreter]
        interpreters.extend(self.other_interpreters)
        environ.update(
            PEX_PYTHON_PATH=os.pathsep.join(interpreter.binary for interpreter in interpreters),
            PEX_TOOLS="1",
        )
        environ.update(env)

        _, stdout, _ = self.interpreter.execute(args=cmd, env=environ)
        return stdout


@pytest.fixture(scope="module")
def interpreter_tool(
    python38,  # type: PythonInterpreter
    python310,  # type: PythonInterpreter
):
    # type: (...) -> InterpreterTool
    return InterpreterTool.create(python38, python310)


def expected_basic(interpreter):
    # type: (PythonInterpreter) -> str
    return interpreter.binary


def test_basic(
    python38,  # type: PythonInterpreter
    interpreter_tool,  # type: InterpreterTool
):
    # type: (...) -> None
    output = interpreter_tool.run()
    assert expected_basic(python38) == output.strip()


def test_basic_all(
    python38,  # type: PythonInterpreter
    python310,  # type: PythonInterpreter
    interpreter_tool,  # type: InterpreterTool
):
    # type: (...) -> None
    output = interpreter_tool.run("-a")
    assert [
        expected_basic(interpreter) for interpreter in (python38, python310)
    ] == output.splitlines()


def expected_verbose(interpreter):
    # type: (PythonInterpreter) -> Dict[str, Any]
    return {
        "path": interpreter.binary,
        "platform": str(interpreter.platform),
        "requirement": str(InterpreterConstraint.exact_version(interpreter)),
    }


def test_verbose(
    python38,  # type: PythonInterpreter
    interpreter_tool,  # type: InterpreterTool
):
    # type: (...) -> None
    output = interpreter_tool.run("-v")
    assert expected_verbose(python38) == json.loads(output)


def test_verbose_all(
    python38,  # type: PythonInterpreter
    python310,  # type: PythonInterpreter
    interpreter_tool,  # type: InterpreterTool
):
    # type: (...) -> None
    output = interpreter_tool.run("-va")
    assert [expected_verbose(interpreter) for interpreter in (python38, python310)] == [
        json.loads(line) for line in output.splitlines()
    ]


def expected_verbose_verbose(interpreter):
    # type: (PythonInterpreter) -> Dict[str, Any]
    expected = expected_verbose(interpreter)
    expected.update(supported_tags=interpreter.identity.supported_tags.to_string_list())
    return expected


def test_verbose_verbose(
    python38,  # type: PythonInterpreter
    interpreter_tool,  # type: InterpreterTool
):
    # type: (...) -> None
    output = interpreter_tool.run("-vv")
    assert expected_verbose_verbose(python38) == json.loads(output)


def test_verbose_verbose_verbose(
    python38,  # type: PythonInterpreter
    interpreter_tool,  # type: InterpreterTool
):
    # type: (...) -> None
    output = interpreter_tool.run("-vvv")
    expected = expected_verbose_verbose(python38)
    expected.update(env_markers=python38.identity.env_markers.as_dict(), venv=False)
    assert expected == json.loads(output)


def test_verbose_verbose_verbose_venv(
    python310,  # type: PythonInterpreter
):
    # type: (...) -> None
    venv = Virtualenv.create(venv_dir=safe_mkdtemp(), interpreter=python310, force=True)
    assert venv.interpreter.is_venv

    # N.B.: Non-venv-mode PEXes always escape venvs to prevent `sys.path` contamination unless
    # `PEX_INHERIT_PATH` is not "false".
    output = InterpreterTool.create(venv.interpreter).run("-vvv", PEX_INHERIT_PATH="fallback")

    expected = expected_verbose_verbose(venv.interpreter)
    expected.update(
        env_markers=venv.interpreter.identity.env_markers.as_dict(),
        venv=True,
        base_interpreter=python310.binary,
    )
    assert expected == json.loads(output)



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex.testing import run_pex_command


def test_pip_2020_resolver_engaged():
    # type: () -> None

    # The Pip legacy resolver cannot solve the following requirements but the 2020 resolver can.
    # Use this fact to prove we're plumbing Pip resolver version arguments correctly.
    pex_args = ["boto3==1.15.6", "botocore>1.17<1.18.7", "--", "-c", "import boto3"]

    results = run_pex_command(args=["--resolver-version", "pip-legacy-resolver"] + pex_args)
    results.assert_failure()
    assert "Failed to resolve compatible distributions:" in results.error
    assert "1: boto3==1.15.6 requires botocore<1.19.0,>=1.18.6 but " in results.error

    run_pex_command(args=["--resolver-version", "pip-2020-resolver"] + pex_args).assert_success()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import subprocess

from pex.build_system import DEFAULT_BUILD_BACKEND
from pex.dist_metadata import Distribution
from pex.interpreter import PythonInterpreter
from pex.pex import PEX
from pex.pex_bootstrapper import VenvPex, ensure_venv
from pex.pex_builder import PEXBuilder
from pex.resolve.resolvers import Resolver
from pex.result import Error
from pex.targets import Target, Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.variables import ENV
from pex.venv.bin_path import BinPath
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Iterable, Mapping, Optional, Tuple, Union

    import attr  # vendor:skip
    import toml  # vendor:skip
else:
    from pex.third_party import attr, toml


@attr.s(frozen=True)
class BuildSystemTable(object):
    requires = attr.ib()  # type: Tuple[str, ...]
    build_backend = attr.ib(default=DEFAULT_BUILD_BACKEND)  # type: str
    backend_path = attr.ib(default=())  # type: Tuple[str, ...]


def _read_build_system_table(
    pyproject_toml,  # type: str
):
    # type: (...) -> Union[Optional[BuildSystemTable], Error]
    try:
        with open(pyproject_toml) as fp:
            data = toml.load(fp)
    except toml.TomlDecodeError as e:
        return Error(
            "Problem parsing toml in {pyproject_toml}: {err}".format(
                pyproject_toml=pyproject_toml, err=e
            )
        )

    build_system = data.get("build-system")
    if not build_system:
        return None

    requires = build_system.get("requires")
    if not requires:
        return None

    return BuildSystemTable(
        requires=tuple(requires),
        build_backend=build_system.get("build-backend", DEFAULT_BUILD_BACKEND),
        backend_path=tuple(
            os.path.join(os.path.dirname(pyproject_toml), entry)
            for entry in build_system.get("backend-path", ())
        ),
    )


@attr.s(frozen=True)
class BuildSystem(object):
    @classmethod
    def create(
        cls,
        interpreter,  # type: PythonInterpreter
        requires,  # type: Iterable[str]
        resolved,  # type: Iterable[Distribution]
        build_backend,  # type: str
        backend_path,  # type: Tuple[str, ...]
        extra_requirements=None,  # type: Optional[Iterable[str]]
        **extra_env  # type: str
    ):
        # type: (...) -> Union[BuildSystem, Error]
        pex_builder = PEXBuilder()
        pex_builder.info.venv = True
        pex_builder.info.venv_site_packages_copies = True
        pex_builder.info.venv_bin_path = BinPath.PREPEND
        for req in requires:
            pex_builder.add_requirement(req)
        for dist in resolved:
            pex_builder.add_distribution(dist)
        pex_builder.freeze(bytecode_compile=False)
        venv_pex = ensure_venv(PEX(pex_builder.path(), interpreter=interpreter))
        if extra_requirements:
            # N.B.: We install extra requirements separately instead of having them resolved and
            # handed in with the `resolved` above because there are cases in the wild where the
            # build system requires (PEP-518) and the results of PEP-517 `get_requires_for_*` can
            # return overlapping requirements. Pip will error for overlaps complaining of duplicate
            # requirements if we attempt to resolve all the requirements at once; so we instead
            # resolve and install in two phases. This obviously has problems! That said, it is, in
            # fact, how Pip's internal PEP-517 build frontend works; so we emulate that.
            virtualenv = Virtualenv(venv_pex.venv_dir)
            # Python 3.5 comes with Pip 9.0.1 which is pretty broken: it doesn't work with our test
            # cases; so we upgrade.
            # For Python 2.7 we use virtualenv (there is no -m venv built into Python) and that
            # comes with Pip 22.0.2, Python 3.6 comes with Pip 18.1 and Python 3.7 comes with
            # Pip 22.04 and the default Pips only get newer with newer version of Pythons. These all
            # work well enough for our test cases and, in general, they should work well enough with
            # the Python they come paired with.
            upgrade_pip = virtualenv.interpreter.version[:2] == (3, 5)
            virtualenv.install_pip(upgrade=upgrade_pip)
            with open(os.devnull, "wb") as dev_null:
                _, process = virtualenv.interpreter.open_process(
                    args=[
                        "-m",
                        "pip",
                        "install",
                        "--ignore-installed",
                        "--no-user",
                        "--no-warn-script-location",
                    ]
                    + list(extra_requirements),
                    stdout=dev_null,
                    stderr=subprocess.PIPE,
                )
                _, stderr = process.communicate()
                if process.returncode != 0:
                    return Error(
                        "Failed to install extra requirement in venv at {venv_dir}: "
                        "{extra_requirements}\nSTDERR:\n{stderr}".format(
                            venv_dir=venv_pex.venv_dir,
                            extra_requirements=", ".join(extra_requirements),
                            stderr=stderr.decode("utf-8"),
                        )
                    )

        # Ensure all PEX* env vars are stripped except for PEX_ROOT and PEX_VERBOSE. We want folks
        # to be able to steer the location of the cache and the logging verbosity, but nothing else.
        # We control the entry-point, etc. of the PEP-518 build backend venv for internal use.
        with ENV.strip().patch(PEX_ROOT=ENV.PEX_ROOT, PEX_VERBOSE=str(ENV.PEX_VERBOSE)) as env:
            if extra_env:
                env.update(extra_env)
            if backend_path:
                env.update(PEX_EXTRA_SYS_PATH=os.pathsep.join(backend_path))
            return cls(
                venv_pex=venv_pex, build_backend=build_backend, requires=tuple(requires), env=env
            )

    venv_pex = attr.ib()  # type: VenvPex
    build_backend = attr.ib()  # type: str
    requires = attr.ib()  # type: Tuple[str, ...]
    env = attr.ib()  # type: Mapping[str, str]


def load_build_system(
    target,  # type: Target
    resolver,  # type: Resolver
    project_directory,  # type: str
    extra_requirements=None,  # type: Optional[Iterable[str]]
):
    # type: (...) -> Union[Optional[BuildSystem], Error]

    # The interface is spec'd here: https://peps.python.org/pep-0518/
    pyproject_toml = os.path.join(project_directory, "pyproject.toml")
    if not os.path.isfile(pyproject_toml):
        return None

    maybe_build_system_table_or_error = _read_build_system_table(pyproject_toml)
    if not isinstance(maybe_build_system_table_or_error, BuildSystemTable):
        return maybe_build_system_table_or_error
    build_system_table = maybe_build_system_table_or_error

    with TRACER.timed(
        "Building {build_backend} build_backend PEX".format(
            build_backend=build_system_table.build_backend
        )
    ):
        result = resolver.resolve_requirements(
            targets=Targets.from_target(target),
            requirements=build_system_table.requires,
        )
        return BuildSystem.create(
            interpreter=target.get_interpreter(),
            requires=build_system_table.requires,
            resolved=tuple(
                installed_distribution.distribution
                for installed_distribution in result.installed_distributions
            ),
            build_backend=build_system_table.build_backend,
            backend_path=build_system_table.backend_path,
            extra_requirements=extra_requirements,
        )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess

import pytest

from pex.interpreter import PythonInterpreter
from pex.testing import (
    ALL_PY_VERSIONS,
    IntegResults,
    VenvFactory,
    all_python_venvs,
    ensure_python_venv,
    make_source_dir,
    run_pex_command,
)
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

    import toml  # vendor:skip
else:
    from pex.third_party import toml


@pytest.mark.parametrize(
    "venv_factory",
    [
        pytest.param(venv_factory, id=venv_factory.python_version)
        for venv_factory in all_python_venvs()
    ],
)
def test_build_isolation(
    venv_factory,  # type: VenvFactory
    pex_project_dir,  # type: str
    tmpdir,  # type: Any
):
    # type: (...) -> None

    python, pip = venv_factory.create_venv()
    result = run_pex_command(args=[pex_project_dir, "--no-build-isolation"], python=python)
    result.assert_failure()
    assert "raise BackendUnavailable(" in result.error, (
        "With build isolation turned off, it's expected that any build requirements (flit for Pex) "
        "are pre-installed. They are not; so we expect a failure here."
    )

    pyproject = toml.load(os.path.join(pex_project_dir, "pyproject.toml"))
    build_requirements = pyproject["build-system"]["requires"]
    assert len(build_requirements) > 0
    subprocess.check_call(args=[pip, "install"] + build_requirements)

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[pex_project_dir, "--no-build-isolation", "-o", pex], python=python
    ).assert_success()
    subprocess.check_call(args=[python, pex, "-c", "import pex"])


def test_pep_517_for_pep_517_project(pex_project_dir):
    # type: (str) -> None

    # N.B.: Pex has a PEP-517 build with no fallback to `setup.py`.

    def build_pex(*extra_args):
        # type: (*str) -> IntegResults
        return run_pex_command(
            args=[pex_project_dir] + list(extra_args) + ["--", "-c", "import pex"]
        )

    build_pex().assert_success()
    build_pex("--force-pep517").assert_success()

    result = build_pex("--no-use-pep517")
    result.assert_failure()
    assert (
        "ERROR: Disabling PEP 517 processing is invalid: project does not have a setup.py"
        in result.error
    )


def test_pep_517_for_legacy_project():
    # type: () -> None

    def assert_build_pex(*extra_args):
        # type: (*str) -> None
        with make_source_dir(name="project", version="0.1.0") as setup_py_project:
            run_pex_command(
                args=[setup_py_project] + list(extra_args) + ["--", "-c", "import project"]
            ).assert_success()

    assert_build_pex()
    assert_build_pex("--use-pep517")
    assert_build_pex("--no-use-pep517")
